#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt/PySide GUI to interface with a a Keysight (former HP or
Agilent) 34970A/34972A data acquisition/switch unit.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
print(__url__)
# pylint: disable=wrong-import-position, missing-function-docstring, bare-except

import sys

import qtpy
from qtpy import QtCore, QtGui, QtWidgets as QtWid
import pyvisa

import dvg_pyqt_controls as controls
from dvg_devices.Keysight_3497xA_protocol_SCPI import Keysight_3497xA
from dvg_devices.Keysight_3497xA_qdev import Keysight_3497xA_qdev

# VISA address of the Keysight 3497xA data acquisition/switch unit containing a
# multiplexer plug-in module. Hence, we simply call this device a 'mux'.
# MUX_VISA_ADDRESS = "USB0::0x0957::0x2007::MY49018071::INSTR"
MUX_VISA_ADDRESS = "GPIB0::9::INSTR"

# SCPI commands to be send to the mux to set up the scan cycle.
"""
scan_list = "(@101:112)"
MUX_SCPI_COMMANDS = [
            f"rout:open {scan_list}",
            f"conf:temp TC,J,{scan_list}",
            f"unit:temp C,{scan_list}",
            f"sens:temp:tran:tc:rjun:type INT,{scan_list}",
            f"sens:temp:tran:tc:check ON,{scan_list}",
            f"sens:temp:nplc 1,{scan_list}",
            f"rout:scan {scan_list}",
]
"""

scan_list = "(@101:110)"
MUX_SCPI_COMMANDS = [
    f"rout:open {scan_list}",
    f"conf:res 1e6,{scan_list}",
    f"sens:res:nplc 1,{scan_list}",
    f"rout:scan {scan_list}",
]

# A scan will be performed by the mux every N milliseconds
DAQ_INTERVAL_MS = 1000  # [ms]

# Show debug info in terminal? Warning: Slow! Do not leave on unintentionally.
DEBUG = False

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(self, qdev: Keysight_3497xA_qdev, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.setWindowTitle("Keysight 3497xA control")
        self.setGeometry(40, 60, 0, 0)
        self.setFont(QtGui.QFont("Arial", 9))
        self.setStyleSheet(
            controls.SS_TEXTBOX_READ_ONLY
            + controls.SS_GROUP
            + controls.SS_HOVER
        )

        # Top grid
        self.qlbl_title = QtWid.QLabel("Keysight 3497xA control")
        self.qlbl_title.setFont(
            QtGui.QFont("Palatino", 14, weight=QtGui.QFont.Weight.Bold)
        )
        self.qpbt_exit = QtWid.QPushButton("Exit")
        self.qpbt_exit.clicked.connect(self.close)
        self.qpbt_exit.setMinimumHeight(30)

        grid_top = QtWid.QGridLayout()
        grid_top.addWidget(self.qlbl_title, 0, 0)
        grid_top.addWidget(
            self.qpbt_exit, 0, 1, QtCore.Qt.AlignmentFlag.AlignRight
        )

        # Bottom grid
        hbox1 = QtWid.QHBoxLayout()
        hbox1.addWidget(qdev.qgrp)
        hbox1.addStretch(1)
        hbox1.setAlignment(qdev.qgrp, QtCore.Qt.AlignmentFlag.AlignTop)

        # Round up full window
        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(grid_top)
        vbox.addLayout(hbox1)
        vbox.addStretch(1)


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":

    # --------------------------------------------------------------------------
    #   Connect to Keysight 3497xA (mux)
    # --------------------------------------------------------------------------

    rm = pyvisa.ResourceManager()
    mux = Keysight_3497xA(MUX_VISA_ADDRESS, "MUX")

    try:
        if mux.connect(rm):
            mux.begin(MUX_SCPI_COMMANDS)
    except ValueError as e:
        # No connection could be made to the VISA device because module
        # dependencies are missing. Print error, not raise and continue to
        # show the GUI.
        print(e)

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------

    main_thread = QtCore.QThread.currentThread()
    if isinstance(main_thread, QtCore.QThread):
        main_thread.setObjectName("MAIN")  # For DEBUG info

    if qtpy.PYQT6 or qtpy.PYSIDE6:
        sys.argv += ["-platform", "windows:darkmode=0"]
    app = QtWid.QApplication(sys.argv)
    app.setStyle("Fusion")

    # --------------------------------------------------------------------------
    #   Set up communication threads for the mux
    # --------------------------------------------------------------------------

    mux_qdev = Keysight_3497xA_qdev(
        dev=mux,
        DAQ_interval_ms=DAQ_INTERVAL_MS,
        debug=DEBUG,
    )
    mux_qdev.start()

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    def about_to_quit():
        print("About to quit")
        app.processEvents()
        mux_qdev.quit()
        try:
            mux.close()
        except:
            pass
        try:
            rm.close()
        except:
            pass

    app.aboutToQuit.connect(about_to_quit)
    window = MainWindow(qdev=mux_qdev)
    window.show()

    sys.exit(app.exec())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt/PySide module to provide multithreaded communication and periodical data
acquisition for MDrive stepper motors by Novanta IMS (former Schneider Electric)
set up in party mode.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "12-06-2024"
__version__ = "1.5.0"
# pylint: disable=broad-except, missing-function-docstring
# pylint: disable=multiple-statements, unnecessary-lambda, too-many-lines

from enum import unique, IntEnum
from functools import partial
from typing import List

from qtpy import QtCore, QtGui, QtWidgets as QtWid
from qtpy.QtCore import Signal, Slot  # type: ignore
import numpy as np

import dvg_pyqt_controls as controls
from dvg_debug_functions import print_fancy_traceback as pft
from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.MDrive_stepper_protocol_RS422 import (
    MDrive_Controller,
    MDrive_Motor,
    Movement_type,
)

FONT_MONOSPACE = QtGui.QFont("Consolas", 12, weight=QtGui.QFont.Weight.Bold)
FONT_MONOSPACE.setStyleHint(QtGui.QFont.StyleHint.TypeWriter)

# fmt: off
SS_PUSHBUTTON_PADDING = (
    "QPushButton {"
        "padding: 8px;}"
)

SS_LED = (
    "QPushButton {"
        "margin: 0;"
        "padding: 0;"
        "border: 1px solid gray;"
        "border-radius: 5px;"
        "max-height: 20px;"
        "max-width: 20px;"
        "height: 20px;"
        "width: 20px;}"
)

SS_LED_RED2GREEN = (
    "QPushButton {"
        "background-color: " + controls.COLOR_ERROR_RED + ";}"
    "QPushButton:checked {"
        "background-color: " + controls.COLOR_LED_GREEN + ";}"
)

SS_LED_NONE2GREEN = (
    "QPushButton {"
        "background-color: transparent;}"
    "QPushButton:checked {"
        "background-color: " + controls.COLOR_LED_GREEN + ";}"
)

SS_ERROR_BOX = (
    "QLineEdit {"
        "background-color: " + controls.COLOR_ERROR_RED + ";}"
    "QLineEdit:read-only {"
        "background-color: transparent;}"
)

SS_GROUPBOX = (
    "QGroupBox {"
        "padding: 0;"
        "margin-left: 0;"
        "margin-top: 3ex;}"
    "QGroupBox::title {"
        "subcontrol-origin: margin;"
        "subcontrol-position: top left;"
        "padding: 0;"
        "margin: 0;"
        "top: 0;"
        "left: 0;}"
    "QGroupBox:flat {"
        "border: 0px;"
        "border-radius: 0 0px;"
        "padding: 0;"
        "margin-left: 0;"
        "margin-top: 3ex;}"
)
# fmt: on


@unique
class GUI_elements(IntEnum):
    ALL = 1
    TAB_CONTROL = 2
    TAB_MOTION = 3
    TAB_DEVICE = 4


# ------------------------------------------------------------------------------
#   GUI_MDrive_motor_panel
# ------------------------------------------------------------------------------


class GUI_MDrive_motor_panel(QtCore.QObject):
    """GUI elements for a single MDrive motor, bundled into a single
    `QtWidgets.QGroupBox` stored in member `main_group_box`."""

    def __init__(
        self,
        controller_qdev: QDeviceIO,  # == `MDrive_Controller_qdev`
        motor: MDrive_Motor,
        **kwargs,
    ):
        super().__init__(**kwargs)  # Pass kwargs onto QtCore.QObject()

        self.controller_qdev = controller_qdev
        self.motor = motor
        self._create_panel()

    def _create_panel(self):
        """Create a GUI panel for a single MDrive motor."""

        # Lists of labels denoting physical units
        self.unit_labels_0: list[QtWid.QLabel] = []  # [mm]     or [rev]
        self.unit_labels_1: list[QtWid.QLabel] = []  # [mm/s]   or [rev/s]
        self.unit_labels_2: list[QtWid.QLabel] = []  # [mm/s^2] or [rev/s^2]

        # Main tab widget of the panel
        self.qtab = QtWid.QTabWidget()

        # ---------------------
        #   Tab page: Control
        # ---------------------

        p1 = {"parent": None, "checkable": True, "enabled": False}
        # fmt: off
        self.led_is_home_known        = QtWid.QPushButton(**p1)
        self.led_is_moving            = QtWid.QPushButton(**p1)
        self.led_is_velocity_changing = QtWid.QPushButton(**p1)
        # fmt: on

        self.led_is_home_known.setStyleSheet(SS_LED + SS_LED_RED2GREEN)
        self.led_is_moving.setStyleSheet(SS_LED + SS_LED_NONE2GREEN)
        self.led_is_velocity_changing.setStyleSheet(SS_LED + SS_LED_NONE2GREEN)

        self.error_status = QtWid.QLineEdit("0")
        self.error_status.setFixedWidth(controls.e8(4))
        self.error_status.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)
        self.error_status.setReadOnly(True)
        self.error_status.setStyleSheet(SS_ERROR_BOX)

        p_mono = {
            "parent": None,
            "font": FONT_MONOSPACE,
            "alignment": QtCore.Qt.AlignmentFlag.AlignTop,
        }
        p1 = {
            **p_mono,
            "alignment": (
                QtCore.Qt.AlignmentFlag.AlignTop
                | QtCore.Qt.AlignmentFlag.AlignRight
            ),
            "minimumWidth": controls.e8(12, FONT_MONOSPACE),
            "textInteractionFlags": (
                QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
            ),
        }
        position_label = QtWid.QLabel(" Position", **p_mono)
        velocity_label = QtWid.QLabel(" Velocity", **p_mono)
        self.position = QtWid.QLabel("nan", **p1)
        self.velocity = QtWid.QLabel("nan", **p1)
        position_unit = QtWid.QLabel(**p_mono)
        velocity_unit = QtWid.QLabel(**p_mono)

        p1 = {
            "parent": None,
            "alignment": QtCore.Qt.AlignmentFlag.AlignRight,
        }
        p2 = {**p1, "maximumWidth": controls.e8(14)}
        # fmt: off
        self.wanted_position         = QtWid.QLineEdit("0.0" , **p1)
        self.wanted_velocity         = QtWid.QLineEdit("0.0" , **p1)
        self.software_min_position   = QtWid.QLineEdit("0.0" , **p2)
        self.software_max_position   = QtWid.QLineEdit("0.0" , **p2)
        self.step_size_1             = QtWid.QLineEdit("1.0" , **p2)
        self.step_size_2             = QtWid.QLineEdit("10.0", **p2)

        wanted_position_unit         = QtWid.QLabel()
        wanted_velocity_unit         = QtWid.QLabel()
        software_max_position_unit   = QtWid.QLabel()
        software_min_position_unit   = QtWid.QLabel()
        step_size_1_unit             = QtWid.QLabel()
        step_size_2_unit             = QtWid.QLabel()

        self.pbtn_init               = QtWid.QPushButton("Init interface")
        self.pbtn_home               = QtWid.QPushButton("Home")
        self.pbtn_move_to_position   = QtWid.QPushButton("Move to position")
        self.pbtn_move_with_velocity = QtWid.QPushButton("Move with velocity")
        self.pbtn_controlled_stop    = QtWid.QPushButton("Controlled stop")
        self.pbtn_step_1_plus        = QtWid.QPushButton("Step +")
        self.pbtn_step_2_plus        = QtWid.QPushButton("Step ++")
        self.pbtn_step_1_minus       = QtWid.QPushButton("Step -")
        self.pbtn_step_2_minus       = QtWid.QPushButton("Step --")
        self.pbtn_STOP               = QtWid.QPushButton("\nEMERGENCY STOP\n")
        # fmt: on

        # Connect GUI signals to slots
        self.pbtn_STOP.clicked.connect(self.process_pbtn_STOP)
        self.pbtn_init.clicked.connect(
            lambda: self.controller_qdev.send(self.motor.init_interface)
        )
        self.pbtn_home.clicked.connect(
            lambda: self.controller_qdev.send(self.motor.home)
        )
        self.pbtn_controlled_stop.clicked.connect(
            lambda: self.controller_qdev.send(self.motor.controlled_stop)
        )
        self.wanted_position.editingFinished.connect(
            lambda: self.validate_qlineedit_float(self.wanted_position)
        )
        self.wanted_velocity.editingFinished.connect(
            lambda: self.validate_qlineedit_float(self.wanted_velocity)
        )
        self.step_size_1.editingFinished.connect(
            lambda: self.validate_qlineedit_float(self.step_size_1)
        )
        self.step_size_2.editingFinished.connect(
            lambda: self.validate_qlineedit_float(self.step_size_2)
        )
        self.pbtn_move_to_position.clicked.connect(
            self.process_pbtn_move_to_position
        )
        self.pbtn_move_with_velocity.clicked.connect(
            self.process_pbtn_move_with_velocity
        )
        self.pbtn_step_1_plus.clicked.connect(
            lambda: self.process_pbtn_step(self.step_size_1, sign_=1)
        )
        self.pbtn_step_2_plus.clicked.connect(
            lambda: self.process_pbtn_step(self.step_size_2, sign_=1)
        )
        self.pbtn_step_1_minus.clicked.connect(
            lambda: self.process_pbtn_step(self.step_size_1, sign_=-1)
        )
        self.pbtn_step_2_minus.clicked.connect(
            lambda: self.process_pbtn_step(self.step_size_2, sign_=-1)
        )

        # Unit labels
        self.unit_labels_0.extend(
            [
                position_unit,
                software_min_position_unit,
                software_max_position_unit,
                wanted_position_unit,
                step_size_1_unit,
                step_size_2_unit,
            ]
        )
        self.unit_labels_1.extend(
            [
                velocity_unit,
                wanted_velocity_unit,
            ]
        )

        pbtns = [
            self.pbtn_init,
            self.pbtn_home,
            self.pbtn_move_to_position,
            self.pbtn_move_with_velocity,
            self.pbtn_controlled_stop,
            self.pbtn_step_1_plus,
            self.pbtn_step_2_plus,
            self.pbtn_step_1_minus,
            self.pbtn_step_2_minus,
        ]
        for pbtn in pbtns:
            pbtn.setStyleSheet(SS_PUSHBUTTON_PADDING)

        # fmt: off
        i = 0
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(2)
        grid.setColumnStretch(2, 1)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        subgrid = QtWid.QGridLayout()
        subgrid.setVerticalSpacing(2)
        subgrid.setColumnStretch(5, 1)
        subgrid.setColumnMinimumWidth(0, controls.e8(16))
        subgrid.setColumnMinimumWidth(3, controls.e8(13))
        subgrid.addWidget(QtWid.QLabel("Moving?")           , 0, 0)
        subgrid.addWidget(self.led_is_moving                , 0, 1)
        subgrid.addItem(QtWid.QSpacerItem(20, 1)            , 0, 2)
        subgrid.addWidget(QtWid.QLabel("Home known?")       , 0, 3)
        subgrid.addWidget(self.led_is_home_known            , 0, 4)
        subgrid.addWidget(QtWid.QLabel("Changing velocity?"), 1, 0)
        subgrid.addWidget(self.led_is_velocity_changing     , 1, 1)
        subgrid.addWidget(QtWid.QLabel("Error status")      , 1, 3)
        subgrid.addWidget(self.error_status                 , 1, 4)

        grid.addItem(subgrid                                , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)                , i, 0)      ; i+=1
        grid.addWidget(position_label                       , i, 0)
        grid.addWidget(self.position                        , i, 1)
        grid.addWidget(position_unit                        , i, 2)      ; i+=1
        grid.addWidget(velocity_label                       , i, 0)
        grid.addWidget(self.velocity                        , i, 1)
        grid.addWidget(velocity_unit                        , i, 2)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)               , i, 0)      ; i+=1

        subgrid = QtWid.QGridLayout()
        subgrid.setVerticalSpacing(2)
        subgrid.addWidget(self.pbtn_init                    , 0, 0)
        subgrid.addWidget(QtWid.QLabel("= Subroutine F1")   , 0, 1, 1, 2)
        subgrid.addWidget(self.pbtn_home                    , 1, 0)
        subgrid.addWidget(QtWid.QLabel("= Subroutine F2")   , 1, 1, 1, 2)
        subgrid.addWidget(self.pbtn_move_to_position        , 2, 0)
        subgrid.addWidget(self.wanted_position              , 2, 1)
        subgrid.addWidget(wanted_position_unit              , 2, 2)
        subgrid.addWidget(self.pbtn_move_with_velocity      , 3, 0)
        subgrid.addWidget(self.wanted_velocity              , 3, 1)
        subgrid.addWidget(wanted_velocity_unit              , 3, 2)
        subgrid.addItem(QtWid.QSpacerItem(1, 12)            , 4, 0)
        subgrid.addWidget(self.pbtn_controlled_stop         , 4, 0, 1, 3)

        actuate_group = QtWid.QGroupBox("Actuate")
        actuate_group.setLayout(subgrid)
        actuate_group.setStyleSheet(SS_GROUPBOX)
        actuate_group.setCheckable(True)
        actuate_group.setChecked(False)
        actuate_group.setFlat(True)

        subgrid = QtWid.QGridLayout()
        subgrid.setVerticalSpacing(2)
        subgrid.setColumnStretch(2, 1)
        subgrid.setColumnMinimumWidth(0, controls.e8(13))
        subgrid.addWidget(QtWid.QLabel("Min position")      , 0, 0)
        subgrid.addWidget(self.software_min_position        , 0, 1)
        subgrid.addWidget(software_min_position_unit        , 0, 2)
        subgrid.addWidget(QtWid.QLabel("Max position")      , 1, 0)
        subgrid.addWidget(self.software_max_position        , 1, 1)
        subgrid.addWidget(software_max_position_unit        , 1, 2)

        limits_group = QtWid.QGroupBox("Software limits")
        limits_group.setLayout(subgrid)
        limits_group.setStyleSheet(SS_GROUPBOX)
        limits_group.setCheckable(True)
        limits_group.setChecked(False)
        limits_group.setFlat(True)

        subsubgrid = QtWid.QGridLayout()
        subsubgrid.setSpacing(0)
        subsubgrid.addWidget(self.pbtn_step_1_minus         , 0, 0)
        subsubgrid.addWidget(self.pbtn_step_1_plus          , 0, 1)
        subsubgrid.addWidget(self.pbtn_step_2_minus         , 1, 0)
        subsubgrid.addWidget(self.pbtn_step_2_plus          , 1, 1)

        subgrid = QtWid.QGridLayout()
        subgrid.setVerticalSpacing(2)
        subgrid.setColumnStretch(2, 1)
        subgrid.setColumnMinimumWidth(0, controls.e8(13))
        subgrid.addWidget(QtWid.QLabel("Step size +/-")     , 0, 0)
        subgrid.addWidget(self.step_size_1                  , 0, 1)
        subgrid.addWidget(step_size_1_unit                  , 0, 2)
        subgrid.addWidget(QtWid.QLabel("Step size ++/--")   , 1, 0)
        subgrid.addWidget(self.step_size_2                  , 1, 1)
        subgrid.addWidget(step_size_2_unit                  , 1, 2)
        subgrid.addItem(QtWid.QSpacerItem(1, 6)             , 2, 0)
        subgrid.addLayout(subsubgrid                        , 3, 0, 1, 3)

        step_group = QtWid.QGroupBox("Step control")
        step_group.setLayout(subgrid)
        step_group.setStyleSheet(SS_GROUPBOX)
        step_group.setCheckable(True)
        step_group.setChecked(False)
        step_group.setFlat(True)

        grid.addWidget(actuate_group                        , i, 0, 1, 3); i+=1
        grid.addWidget(limits_group                         , i, 0, 1, 3); i+=1
        grid.addWidget(step_group                           , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 12)               , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_STOP                       , i, 0, 1, 3); i+=1
        # fmt: on

        # Disable `limits_group` because functionality is not implemented yet.
        # TODO: Implement. Might need to add members `soft_min_position` and
        # `soft_max_position` to `MDrive_Motor.config`.
        limits_group.setVisible(False)

        grid_widget = QtWid.QWidget()
        grid_widget.setLayout(grid)
        self.page_control = self.qtab.addTab(grid_widget, "Control")

        # ---------------------
        #   Tab page: Motion
        # ---------------------

        def add_motion_row(
            grid: QtWid.QGridLayout,
            row_idx: List[int],  # Single int put in a list to allow pass by ref
            descr: str = "",
            abbrev: str = "",
            unit: str = "",
        ):

            unit_label = QtWid.QLabel(unit)
            line_edit = QtWid.QLineEdit()
            line_edit.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
            line_edit.setFixedWidth(controls.e8(10))
            grid.addWidget(QtWid.QLabel(descr), row_idx[0], 0)
            grid.addWidget(QtWid.QLabel(abbrev), row_idx[0], 1)
            grid.addWidget(line_edit, row_idx[0], 2)
            grid.addWidget(unit_label, row_idx[0], 3)
            row_idx[0] += 1

            return line_edit, unit_label

        def add_spacer_row(grid: QtWid.QGridLayout, row_idx: List[int]):
            grid.addItem(QtWid.QSpacerItem(1, 10), row_idx[0], 0)
            row_idx[0] += 1

        self.cmbx_movement_type = QtWid.QComboBox()
        self.cmbx_movement_type.addItems([type.name for type in Movement_type])

        # fmt: off
        row_idx = [0]
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(2)
        grid.setColumnStretch(3, 1)
        grid.setColumnMinimumWidth(3, controls.e8(10))
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        grid.addWidget(QtWid.QLabel("Movement type"), row_idx[0], 0)
        grid.addWidget(QtWid.QLabel("CT")           , row_idx[0], 1)
        grid.addWidget(self.cmbx_movement_type      , row_idx[0], 2, 1, 2)
        row_idx[0] += 1
        # fmt: on

        (
            self.calibration,
            self.calibration_unit,
        ) = add_motion_row(grid, row_idx, "Calibration", "C0", "")
        add_spacer_row(grid, row_idx)
        (
            self.acceleration_steps,
            _,
        ) = add_motion_row(grid, row_idx, "Acceleration", "A", "steps/sec^2")
        (
            self.deceleration_steps,
            _,
        ) = add_motion_row(grid, row_idx, "Deceleration", "D", "steps/sec^2")
        (
            self.initial_velocity_steps,
            _,
        ) = add_motion_row(grid, row_idx, "Initial velocity", "VI", "steps/sec")
        (
            self.maximum_velocity_steps,
            _,
        ) = add_motion_row(grid, row_idx, "Maximum velocity", "VM", "steps/sec")
        add_spacer_row(grid, row_idx)
        (
            self.acceleration,
            acceleration_unit,
        ) = add_motion_row(grid, row_idx, "Acceleration")
        (
            self.deceleration,
            deceleration_unit,
        ) = add_motion_row(grid, row_idx, "Deceleration")
        (
            self.initial_velocity,
            initial_velocity_unit,
        ) = add_motion_row(grid, row_idx, "Initial velocity")
        (
            self.maximum_velocity,
            maximum_velocity_unit,
        ) = add_motion_row(grid, row_idx, "Maximum velocity")
        add_spacer_row(grid, row_idx)
        (
            self.microsteps,
            _,
        ) = add_motion_row(grid, row_idx, "Microsteps", "MS", "microsteps")
        (
            self.limit_stop,
            _,
        ) = add_motion_row(grid, row_idx, "Limit stop", "LM", "mode 1-6")
        (
            self.run_current,
            _,
        ) = add_motion_row(grid, row_idx, "Run current", "RC", "%")
        (
            self.hold_current,
            _,
        ) = add_motion_row(grid, row_idx, "Hold current", "HC", "%")
        (
            self.hold_delay,
            _,
        ) = add_motion_row(grid, row_idx, "Hold delay", "HT", "msec")
        (
            self.settling_delay,
            _,
        ) = add_motion_row(grid, row_idx, "Settling delay", "MT", "msec")
        add_spacer_row(grid, row_idx)
        (
            self.IO_S1,
            _,
        ) = add_motion_row(grid, row_idx, "I/O point S1", "S1", "#, #, #")
        (
            self.IO_S2,
            _,
        ) = add_motion_row(grid, row_idx, "I/O point S2", "S2", "#, #, #")
        (
            self.IO_S3,
            _,
        ) = add_motion_row(grid, row_idx, "I/O point S3", "S3", "#, #, #")
        (
            self.IO_S4,
            _,
        ) = add_motion_row(grid, row_idx, "I/O point S4", "S4", "#, #, #")
        add_spacer_row(grid, row_idx)

        self.pbtn_save_to_NVM = QtWid.QPushButton(
            "Save to MDrive non-volatile memory"
        )
        self.pbtn_save_to_NVM.setStyleSheet(SS_PUSHBUTTON_PADDING)
        self.pbtn_save_to_NVM.clicked.connect(
            lambda: self.controller_qdev.send(self.motor.save_to_NVM)
        )

        # fmt: off
        row_idx[0] += 1
        grid.addWidget(
            QtWid.QLabel(
                "NOTE: Changes to above parameters are in effect\n"
                "immediately, but don't persist a power-down/up\n"
                "cycle unless saved to MDrive non-volatile memory."
            )                                           , row_idx[0], 0, 1, 4)
        row_idx[0] += 1
        grid.addItem(QtWid.QSpacerItem(1, 6)            , row_idx[0], 0)
        row_idx[0] += 1
        grid.addWidget(self.pbtn_save_to_NVM            , row_idx[0], 0, 1, 4)
        # fmt: on

        # Connect GUI signals to slots
        self.cmbx_movement_type.currentIndexChanged.connect(
            lambda x: self.process_cmbx_movement_type(x)
        )
        self.calibration.editingFinished.connect(self.process_calibration)
        self.acceleration_steps.editingFinished.connect(
            lambda: self.process_motion_param("A", self.acceleration_steps)
        )
        self.deceleration_steps.editingFinished.connect(
            lambda: self.process_motion_param("D", self.deceleration_steps)
        )
        self.initial_velocity_steps.editingFinished.connect(
            lambda: self.process_motion_param("VI", self.initial_velocity_steps)
        )
        self.maximum_velocity_steps.editingFinished.connect(
            lambda: self.process_motion_param("VM", self.maximum_velocity_steps)
        )
        self.acceleration.editingFinished.connect(
            lambda: self.process_motion_param_calibrated("A", self.acceleration)
        )
        self.deceleration.editingFinished.connect(
            lambda: self.process_motion_param_calibrated("D", self.deceleration)
        )
        self.initial_velocity.editingFinished.connect(
            lambda: self.process_motion_param_calibrated(
                "VI", self.initial_velocity
            )
        )
        self.maximum_velocity.editingFinished.connect(
            lambda: self.process_motion_param_calibrated(
                "VM", self.maximum_velocity
            )
        )
        self.microsteps.editingFinished.connect(
            lambda: self.process_motion_param("MS", self.microsteps)
        )
        self.limit_stop.editingFinished.connect(
            lambda: self.process_motion_param("LM", self.limit_stop)
        )
        self.run_current.editingFinished.connect(
            lambda: self.process_motion_param("RC", self.run_current)
        )
        self.hold_current.editingFinished.connect(
            lambda: self.process_motion_param("HC", self.hold_current)
        )
        self.hold_delay.editingFinished.connect(
            lambda: self.process_motion_param("HT", self.hold_delay)
        )
        self.settling_delay.editingFinished.connect(
            lambda: self.process_motion_param("MT", self.settling_delay)
        )
        self.IO_S1.editingFinished.connect(
            lambda: self.process_IO_S("S1", self.IO_S1)
        )
        self.IO_S2.editingFinished.connect(
            lambda: self.process_IO_S("S2", self.IO_S2)
        )
        self.IO_S3.editingFinished.connect(
            lambda: self.process_IO_S("S3", self.IO_S3)
        )
        self.IO_S4.editingFinished.connect(
            lambda: self.process_IO_S("S4", self.IO_S4)
        )

        # Unit labels
        self.unit_labels_1.extend(
            [
                initial_velocity_unit,
                maximum_velocity_unit,
            ]
        )

        self.unit_labels_2.extend(
            [
                acceleration_unit,
                deceleration_unit,
            ]
        )

        grid_widget = QtWid.QWidget()
        grid_widget.setLayout(grid)
        self.page_motion = self.qtab.addTab(grid_widget, "Motion params")

        # ---------------------
        #   Tab page: Device
        # ---------------------

        p1 = {
            "minimumWidth": controls.e8(12),
            "readOnly": True,
        }
        p2 = {
            **p1,
            "alignment": QtCore.Qt.AlignmentFlag.AlignLeft,
        }

        self.device_name = QtWid.QLineEdit(**p2)
        self.device_name.setFixedWidth(controls.e8(2))
        self.part_number = QtWid.QLineEdit(**p2)
        self.serial_number = QtWid.QLineEdit(**p2)
        self.firmware_version = QtWid.QLineEdit(**p2)
        self.user_subroutines = QtWid.QTextEdit(**p1)
        self.user_subroutines.setFont(FONT_MONOSPACE)
        self.user_subroutines.setFontPointSize(10)
        self.user_variables = QtWid.QTextEdit(**p1)
        self.user_variables.setFont(FONT_MONOSPACE)
        self.user_variables.setFontPointSize(10)

        i = 0
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(2)
        grid.setColumnStretch(2, 1)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # fmt: off
        grid.addWidget(QtWid.QLabel("<b>Device</b>")   , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Device name")     , i, 0)
        grid.addWidget(QtWid.QLabel("DN")              , i, 1)
        grid.addWidget(self.device_name                , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Part number")     , i, 0)
        grid.addWidget(QtWid.QLabel("PN")              , i, 1)
        grid.addWidget(self.part_number                , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Serial number")   , i, 0)
        grid.addWidget(QtWid.QLabel("SN")              , i, 1)
        grid.addWidget(self.serial_number              , i, 2)      ; i+=1
        grid.addWidget(QtWid.QLabel("Firmware version"), i, 0)
        grid.addWidget(QtWid.QLabel("VR")              , i, 1)
        grid.addWidget(self.firmware_version           , i, 2)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1

        grid.addWidget(QtWid.QLabel("<b>User subroutines</b>")
                                                       , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1
        grid.addWidget(self.user_subroutines           , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1

        grid.addWidget(QtWid.QLabel("<b>User variables</b>")
                                                       , i, 0, 1, 3); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 6)           , i, 0)      ; i+=1
        grid.addWidget(self.user_variables             , i, 0, 1, 3); i+=1
        # fmt: on

        grid_widget = QtWid.QWidget()
        grid_widget.setLayout(grid)
        self.page_device = self.qtab.addTab(grid_widget, "Device")

        # ---------------
        #   Rounding up
        # ---------------

        hbox = QtWid.QHBoxLayout()
        hbox.addWidget(self.qtab)

        self.main_group_box = QtWid.QGroupBox(f"Motor {self.motor.device_name}")
        self.main_group_box.setLayout(hbox)

    # --------------------------------------------------------------------------
    #   update_GUI
    # --------------------------------------------------------------------------

    @Slot()
    @Slot(GUI_elements)
    def update_GUI(self, gui_elements: GUI_elements = GUI_elements.ALL):

        if not self.motor.controller.is_alive:
            # TODO: Signal a 'lost connection to motor' to the user
            return

        # Shorthands
        state = self.motor.state
        config = self.motor.config
        C0 = config.calibration_constant

        if gui_elements in [GUI_elements.ALL, GUI_elements.TAB_CONTROL]:
            # fmt: off
            self.led_is_home_known       .setChecked(state.is_home_known)
            self.led_is_moving           .setChecked(state.is_moving)
            self.led_is_velocity_changing.setChecked(state.is_velocity_changing)
            self.error_status            .setText(f"{state.error}")
            self.error_status            .setReadOnly(not state.has_error)

            self.position.setText(f"{state.position / C0:.3f}")
            self.velocity.setText(f"{state.velocity / C0:.3f}")
            # fmt: on

        if gui_elements in [GUI_elements.ALL, GUI_elements.TAB_MOTION]:
            self.update_unit_labels(self.motor.get_base_unit())
            self.cmbx_movement_type.setCurrentIndex(config.movement_type.value)
            # fmt: off
            self.calibration           .setText(f"{C0}")
            self.acceleration_steps    .setText(f"{config.motion_A      :.0f}")
            self.deceleration_steps    .setText(f"{config.motion_D      :.0f}")
            self.initial_velocity_steps.setText(f"{config.motion_VI     :.0f}")
            self.maximum_velocity_steps.setText(f"{config.motion_VM     :.0f}")
            self.acceleration          .setText(f"{config.motion_A  / C0:.4f}")
            self.deceleration          .setText(f"{config.motion_D  / C0:.4f}")
            self.initial_velocity      .setText(f"{config.motion_VI / C0:.4f}")
            self.maximum_velocity      .setText(f"{config.motion_VM / C0:.4f}")
            self.microsteps            .setText(f"{config.motion_MS}")
            self.limit_stop            .setText(f"{config.motion_LM}")
            self.run_current           .setText(f"{config.motion_RC}")
            self.hold_current          .setText(f"{config.motion_HC}")
            self.hold_delay            .setText(f"{config.motion_HT}")
            self.settling_delay        .setText(f"{config.motion_MT}")
            self.IO_S1                 .setText(f"{config.IO_S1}")
            self.IO_S2                 .setText(f"{config.IO_S2}")
            self.IO_S3                 .setText(f"{config.IO_S3}")
            self.IO_S4                 .setText(f"{config.IO_S4}")
            # fmt: on

            lines: list[str] = []
            for _idx, (key, val) in enumerate(config.user_subroutines.items()):
                lines.append(f"{key} @ {val}")
            str_text = "\n".join(lines)
            self.user_subroutines.setText(str_text)

            lines = []
            for _idx, (key, val) in enumerate(config.user_variables.items()):
                lines.append(f"{key} = {val}")
            str_text = "\n".join(lines)
            self.user_variables.setText(str_text)

        if gui_elements in [GUI_elements.ALL, GUI_elements.TAB_DEVICE]:
            # fmt: off
            self.part_number     .setText(f"{config.part_number}")
            self.serial_number   .setText(f"{config.serial_number}")
            self.firmware_version.setText(f"{config.firmware_version}")
            self.device_name     .setText(f"{self.motor.device_name}")
            # fmt: on

        # locker.unlock()

    # --------------------------------------------------------------------------
    #   Handle controls
    # --------------------------------------------------------------------------

    def update_unit_labels(self, base_unit: str = "mm"):
        self.calibration_unit.setText(f"steps/{base_unit}")

        for unit_label in self.unit_labels_0:
            unit_label.setText(base_unit)

        for unit_label in self.unit_labels_1:
            unit_label.setText(f"{base_unit}/sec")

        for unit_label in self.unit_labels_2:
            unit_label.setText(f"{base_unit}/sec^2")

    def validate_qlineedit_float(
        self, qlin: QtWid.QLineEdit, default_val: float = 0.0
    ):
        try:
            val = float(qlin.text())
        except ValueError:
            val = default_val
        except Exception as e:
            raise e

        qlin.setText(f"{val}")

    @Slot()
    def process_pbtn_STOP(self):
        self.controller_qdev.add_to_jobs_queue(self.motor.controller.STOP)
        self.controller_qdev.add_to_jobs_queue(self.motor.controller.RESET)
        self.controller_qdev.process_jobs_queue()

    @Slot()
    def process_pbtn_move_to_position(self):
        try:
            val = float(self.wanted_position.text())
        except ValueError:
            val = 0.0
        except Exception as e:
            raise e

        if self.motor.config.movement_type == Movement_type.LINEAR:
            fun = self.motor.move_absolute_mm
        else:
            fun = self.motor.move_absolute_rev

        self.controller_qdev.send(fun, val)

    @Slot()
    def process_pbtn_move_with_velocity(self):
        try:
            val = float(self.wanted_velocity.text())
        except ValueError:
            val = 0.0
        except Exception as e:
            raise e

        if self.motor.config.movement_type == Movement_type.LINEAR:
            fun = self.motor.slew_mm_per_sec
        else:
            fun = self.motor.slew_rev_per_sec

        self.controller_qdev.send(fun, val)

    def process_pbtn_step(self, qlin: QtWid.QLineEdit, sign_: int = 1):
        try:
            val = float(qlin.text())
        except ValueError:
            val = 0.0
        except Exception as e:
            raise e

        if self.motor.config.movement_type == Movement_type.LINEAR:
            fun = self.motor.move_relative_mm
        else:
            fun = self.motor.move_relative_rev

        self.controller_qdev.send(fun, val if sign_ == 1 else -val)

    @Slot(int)
    def process_cmbx_movement_type(self, cmbx_index: int):
        try:
            CT = Movement_type(cmbx_index)
        except ValueError:
            CT = Movement_type.LINEAR
        except Exception as e:
            raise e

        if self.motor.config.movement_type == CT:
            # Prevent race condition
            return

        cmd = f"CT {CT.value:d}"
        self.controller_qdev.add_to_jobs_queue(self.motor.query, cmd)
        self.controller_qdev.add_to_jobs_queue(self.motor.query_config)
        self.controller_qdev.add_to_jobs_queue(
            "signal_GUI_update", (self, GUI_elements.TAB_MOTION)
        )
        self.controller_qdev.process_jobs_queue()

    @Slot()
    def process_calibration(self):
        try:
            C0 = int(self.calibration.text())
            C0 = max(1, C0)
        except ValueError:
            C0 = self.motor.config.calibration_constant
        except Exception as e:
            raise e

        cmd = f"C0 {C0:d}"
        self.controller_qdev.add_to_jobs_queue(self.motor.query, cmd)
        self.controller_qdev.add_to_jobs_queue(self.motor.query_config)
        self.controller_qdev.add_to_jobs_queue(
            "signal_GUI_update", (self, GUI_elements.TAB_MOTION)
        )
        self.controller_qdev.process_jobs_queue()

    def process_motion_param(
        self,
        motion_param: str,
        qlin: QtWid.QLineEdit,
    ):
        """Cast the text string of the passed QLineEdit `qlin` into an integer
        and send this value to the MDrive as a new `motion_param`. Afterwards,
        it will trigger a GUI update of the "Motion" tab."""
        try:
            val = int(qlin.text())
        except ValueError:
            val = 0
        except Exception as e:
            raise e

        cmd = f"{motion_param} {val:d}"
        self.controller_qdev.add_to_jobs_queue(self.motor.query, cmd)
        self.controller_qdev.add_to_jobs_queue(self.motor.query_config)
        self.controller_qdev.add_to_jobs_queue(
            "signal_GUI_update", (self, GUI_elements.TAB_MOTION)
        )
        self.controller_qdev.process_jobs_queue()

    def process_motion_param_calibrated(
        self,
        motion_param: str,
        qlin: QtWid.QLineEdit,
    ):
        """Cast the text string of the passed QLineEdit `qlin` into a float,
        multiply it with the calibration constant C0 and send this rounded value
        to the MDrive as a new `motion_param`. Afterwards, it will trigger a GUI
        update of the "Motion" tab."""
        try:
            val = float(qlin.text())
        except ValueError:
            val = 0
        except Exception as e:
            raise e

        C0 = self.motor.config.calibration_constant
        val = int(round(val * (C0 if not np.isnan(C0) else 1)))

        cmd = f"{motion_param} {val:d}"
        self.controller_qdev.add_to_jobs_queue(self.motor.query, cmd)
        self.controller_qdev.add_to_jobs_queue(self.motor.query_config)
        self.controller_qdev.add_to_jobs_queue(
            "signal_GUI_update", (self, GUI_elements.TAB_MOTION)
        )
        self.controller_qdev.process_jobs_queue()

    def process_IO_S(
        self,
        motion_param: str,
        qlin: QtWid.QLineEdit,
    ):
        """Concerns motion parameters "S1" to "S4" and "S9" to "S12". Validate
        the text string of the passed QLineEdit `qlin`. Expecting the following
        format: "[0-23], [0/1], [0/1]", e.g. "3, 0, 1". The validated string
        will be send to the MDrive as a new `motion_param`. Afterwards, it will
        trigger a GUI update of the "Motion" tab."""
        values: list[int] = [0, 0, 0]

        # Expecting the following format: [0-23], [0/1], [0/1]
        text_parts = qlin.text().split(",")
        for idx, text_part in enumerate(text_parts):
            if idx > 2:
                break

            try:
                val = int(text_part.strip())
            except ValueError:
                val = 0
            except Exception as e:
                raise e

            """ We'll leave number range validation to the MDrive motor
            if idx == 0:
                val = max(val, 0)
                val = min(val, 23)
            """

            if idx in (1, 2):
                val = max(val, 0)
                val = min(val, 1)

            values[idx] = val

        validated_text = f"{values[0]}, {values[1]}, {values[2]}"
        qlin.setText(validated_text)

        if validated_text == self.motor.config.IO_S1:
            return

        cmd = f"{motion_param} {validated_text}"
        self.controller_qdev.add_to_jobs_queue(self.motor.query, cmd)
        self.controller_qdev.add_to_jobs_queue(self.motor.query_config)
        self.controller_qdev.add_to_jobs_queue(
            "signal_GUI_update", (self, GUI_elements.TAB_MOTION)
        )
        self.controller_qdev.process_jobs_queue()


# ------------------------------------------------------------------------------
#   MDrive_Controller_qdev
# ------------------------------------------------------------------------------


class MDrive_Controller_qdev(QDeviceIO):
    """TODO: docstr"""

    signal_GUI_update = Signal(GUI_MDrive_motor_panel, GUI_elements)
    """Triggers updating GUI elements."""

    def __init__(
        self,
        dev: MDrive_Controller,
        DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
        DAQ_interval_ms=200,
        DAQ_timer_type=QtCore.Qt.TimerType.CoarseTimer,
        critical_not_alive_count=3,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()
        self.dev: MDrive_Controller  # Enforce type: removes `_NoDevice()`

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_trigger,
            DAQ_function=self.DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            debug=debug,
        )

        self.create_worker_jobs(jobs_function=self.jobs_function, debug=debug)

        self.motor_panels: list[GUI_MDrive_motor_panel] = []
        """List of MDrive motor panels, one panel for each motor."""

        self.hbox = QtWid.QHBoxLayout()
        """Main `QtWidgets.QHBoxLayout` containing all MDrive motor panels."""

        # Create panels for all MDrive motors
        for motor in self.dev.motors:
            motor_panel = GUI_MDrive_motor_panel(
                controller_qdev=self, motor=motor
            )
            self.motor_panels.append(motor_panel)

            # Each motor panel has its GUI controls grouped together in a single
            # QGroupBox, located in member `main_group_box`.
            self.hbox.addWidget(motor_panel.main_group_box)

            # Trigger an update of the 'control' tab GUI elements whenever the
            # DAQ updated, i.e. new readings are available which should get
            # displayed.
            self.signal_DAQ_updated.connect(
                partial(motor_panel.update_GUI, GUI_elements.TAB_CONTROL)
            )

        # Listen to requests to update GUI elements
        self.signal_GUI_update.connect(
            lambda motor_panel, elements: motor_panel.update_GUI(elements)
        )

        for motor_panel in self.motor_panels:
            motor_panel.update_GUI()

    # --------------------------------------------------------------------------
    #   DAQ_function
    # --------------------------------------------------------------------------

    def DAQ_function(self) -> bool:
        success = True
        for motor in self.dev.motors:
            success &= motor.query_state()
            success &= motor.query_errors()

        return success

    # --------------------------------------------------------------------------
    #   jobs_function
    # --------------------------------------------------------------------------

    def jobs_function(self, func, args):
        if func == "signal_GUI_update":
            # Special instruction
            self.signal_GUI_update.emit(*args)
        else:
            # Default job processing:
            # Send I/O operation to the device
            try:
                func(*args)
            except Exception as err:
                pft(err)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Modbus RTU protocol (RS485) function library for a Xylem Hydrovar HVL
variable speed pump controller.

This module supports just one slave device on the Modbus, not multiple.

When this module is directly run from the terminal a demo will be shown.

Reference documents:
(1) Hydrovar HVL 2.015 - 4.220 | Modbus Protocol & Parameters
    HVL Software Version: 2.10, 2.20
    cod. 001085110 rev. B ed. 01/2018
(2) Hydrovar HVL 2.015 - 4.220 | Installation, Operation, and Maintenance Manual
    cod. 001085102 rev. A ed. 01/2016
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "27-06-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring, too-many-lines

import sys
from enum import IntEnum
import time
from typing import Union, Tuple

import numpy as np

from dvg_debug_functions import print_fancy_traceback as pft
from dvg_devices.BaseDevice import SerialDevice
from dvg_devices.XylemHydrovarHVL_CRC_tools import pretty_hex, crc16

# ------------------------------------------------------------------------------
#   accurate_delay_ms
# ------------------------------------------------------------------------------


def accurate_delay_ms(delay):
    """Accurate time delay in milliseconds, useful for delays < 10 ms.
    The standard `time.sleep()` will not work reliably for such small time
    delays and usually has a minimum delay of up to 10 to 20 ms depending
    on the OS.
    """
    _ = time.perf_counter() + delay / 1000
    while time.perf_counter() < _:
        pass


# ------------------------------------------------------------------------------
#   Enumerations
# ------------------------------------------------------------------------------


class HVL_Mode(IntEnum):
    CONTROLLER = 0  # Regulate motor frequency to maintain a pressure setpoint
    ACTUATOR = 3  # Fixed motor frequency


class HVL_FuncCode(IntEnum):
    # Implemented:
    READ = 0x03  # Read the contents of a single register
    WRITE = 0x06  # Write a value into a single register

    # Not implemented:
    WRITE_CONT = 0x10  # Write values into a block of contiguous registers


class HVL_DType(IntEnum):
    U08 = 0
    U16 = 1
    U32 = 2
    S08 = 3
    S16 = 4
    B0 = 5  # 8 bits bitmap
    B1 = 6  # 16 bits bitmap
    B2 = 7  # 32 bits bitmap


# ------------------------------------------------------------------------------
#   HVL_Register
# ------------------------------------------------------------------------------


class HVL_Register:
    """Placeholder for a single Modbus register adress and its datum type. To be
    populated from Table 5, ref (1).
    """

    def __init__(self, address: int, datum_type: HVL_DType):
        self.address = address
        self.datum_type = datum_type


# List of registers (incomplete list, just the bare necessities).
# Taken from Table 5, ref. (1)
# fmt: off

# M00: Main menu
HVLREG_STOP_START    = HVL_Register(0x0031, HVL_DType.U08)  # RW    P-
HVLREG_ACTUAL_VALUE  = HVL_Register(0x0032, HVL_DType.S16)  # R     P-
HVLREG_EFF_REQ_VAL   = HVL_Register(0x0037, HVL_DType.U16)  # R     P03
HVLREG_START_VALUE   = HVL_Register(0x0038, HVL_DType.U08)  # RW    P04

# M20: Status
HVLREG_ENABLE_DEVICE = HVL_Register(0x0061, HVL_DType.U08)  # RW    P24

# M40: Diagnostics
HVLREG_TEMP_INVERTER = HVL_Register(0x0085, HVL_DType.S08)  # R     P43
HVLREG_CURR_INVERTER = HVL_Register(0x0087, HVL_DType.U16)  # R     P44
HVLREG_VOLT_INVERTER = HVL_Register(0x0088, HVL_DType.U16)  # R     P45
HVLREG_OUTPUT_FREQ   = HVL_Register(0x0033, HVL_DType.S16)  # R     P46

# M100: Basic settings
HVLREG_MODE          = HVL_Register(0x008b, HVL_DType.U08)  # RW    P105

# M200: Conf. inverter
HVLREG_MAX_FREQ      = HVL_Register(0x009d, HVL_DType.U16)  # RW    P245
HVLREG_MIN_FREQ      = HVL_Register(0x009e, HVL_DType.U16)  # RW    P250
HVLREG_MOTOR_NOM_CURR= HVL_Register(0x011a, HVL_DType.U32)  # R     P268

# M600: Error
HVLREG_ERROR_RESET   = HVL_Register(0x00d3, HVL_DType.U08)  # RW    P615

# M800: Required values
HVLREG_C_REQ_VAL_1   = HVL_Register(0x00e5, HVL_DType.U08)  # RW    P805
HVLREG_C_REQ_VAL_2   = HVL_Register(0x00e6, HVL_DType.U08)  # RW    P810
HVLREG_SW_REQ_VAL    = HVL_Register(0x00e7, HVL_DType.U08)  # RW    P815
HVLREG_REQ_VAL_1     = HVL_Register(0x00e8, HVL_DType.U16)  # RW    P820
HVLREG_ACTUAT_FREQ_1 = HVL_Register(0x00ea, HVL_DType.U16)  # RW    P830

# M1000: Test run
HVLREG_TEST_RUN      = HVL_Register(0x00f9, HVL_DType.U08)  # RW    P1005

# M1200: RS-485 Interface
HVLREG_ADDRESS       = HVL_Register(0x010d, HVL_DType.U08)  # RW    P1205

# Special status bits
HVLREG_ERRORS_H3     = HVL_Register(0x012d, HVL_DType.B2)   # R     P-
HVLREG_DEV_STATUS_H4 = HVL_Register(0x01c1, HVL_DType.B1)   # R     P-

# fmt: on

# ------------------------------------------------------------------------------
#   XylemHydrovarHVL
# ------------------------------------------------------------------------------


class XylemHydrovarHVL(SerialDevice):
    class State:
        """Container for the process and measurement variables"""

        # fmt: off
        pump_is_on         = False   # (bool)   Excludes coasting down
        pump_is_running    = False   # (bool)   True with any physical movement
        pump_is_enabled    = False   # (bool)   Inverter input terminal 18
        hvl_mode           = HVL_Mode.CONTROLLER  # (HVL_Mode)          P105

        actual_pressure    = np.nan  # [bar]
        wanted_pressure    = 0.0     # [bar]                            P820

        actual_frequency   = np.nan  # [Hz]                             P46
        wanted_frequency   = 0.0     # [Hz]                             P830

        min_frequency      = np.nan  # [Hz]                             P250
        max_frequency      = np.nan  # [Hz]                             P245
        nom_motor_current  = np.nan  # [A]                              P268

        inverter_temp      = np.nan  # ['C]                             P43
        inverter_volt      = np.nan  # [V]                              P45
        inverter_curr_A    = np.nan  # [A]                              P44
        inverter_curr_pct  = np.nan  # [% FS]     derived: P44 / P268 * 100
        # fmt: on

    class ErrorStatus:
        """Container for the Error Status bits (H3)"""

        # fmt: off
        overcurrent       = False  # bit 00, error 11
        overload          = False  # bit 01, error 12
        overvoltage       = False  # bit 02, error 13
        phase_loss        = False  # bit 03, error 16
        inverter_overheat = False  # bit 04, error 14
        motor_overheat    = False  # bit 05, error 15
        lack_of_water     = False  # bit 06, error 21
        minimum_threshold = False  # bit 07, error 22
        act_val_sensor_1  = False  # bit 08, error 23
        act_val_sensor_2  = False  # bit 09, error 24
        setpoint_1_low_mA = False  # bit 10, error 25
        setpoint_2_low_mA = False  # bit 11, error 26
        # fmt: on

        def has_error(self) -> bool:
            error_list = (
                self.overcurrent,
                self.overload,
                self.overvoltage,
                self.phase_loss,
                self.inverter_overheat,
                self.motor_overheat,
                self.lack_of_water,
                self.minimum_threshold,
                self.act_val_sensor_1,
                self.act_val_sensor_2,
                self.setpoint_1_low_mA,
                self.setpoint_2_low_mA,
            )

            return bool(np.any(error_list))

        def report(self):
            """Report the last read error status to the terminal."""

            if not self.has_error():
                print("No errors")
                return

            print("ERRORS DETECTED")
            print("---------------")
            if self.overcurrent:
                print("- #11: OVERCURRENT")
            if self.overload:
                print("- #12: OVERLOAD")
            if self.overvoltage:
                print("- #13: OVERVOLTAGE")
            if self.phase_loss:
                print("- #16: PHASE LOSS")
            if self.inverter_overheat:
                print("- #14: INVERTER OVERHEAT")
            if self.motor_overheat:
                print("- #15: MOTOR OVERHEAT")
            if self.lack_of_water:
                print("- #21: LACK OF WATER")
            if self.minimum_threshold:
                print("- #22: MINIMUM THRESHOLD")
            if self.act_val_sensor_1:
                print("- #23: ACT VAL SENSOR 1")
            if self.act_val_sensor_2:
                print("- #24: ACT VAL SENSOR 2")
            if self.setpoint_1_low_mA:
                print("- #25: SETPOINT 1 I<4 mA")
            if self.setpoint_2_low_mA:
                print("- #26: SETPOINT 2 I<4 mA")

    class DeviceStatus:
        """Container for the Extended Device Status bits (H4).

        Explanation
        -----------
        external_ON_OFF_terminal_enabled:
          Directly linked to the inverter input terminal 18 'ON_OFF'. Will in
          effect enable or disable running the pump.

        device_is_enabled_with_start_button:
          Gets set by ModBus commands 'pump stop' and 'pump start'. Immediate
          response.

        motor_is_running:
          Autonomously set by the inverter as long as movement of the motor
          drive shaft is detected. Includes coasting up and down.
        """

        # fmt: off
        device_is_preset                    = False  # bit 00
        device_is_ready_for_regulation      = False  # bit 01
        device_has_an_error                 = False  # bit 02
        device_has_a_warning                = False  # bit 03
        external_ON_OFF_terminal_enabled    = False  # bit 04
        device_is_enabled_with_start_button = False  # bit 05
        motor_is_running                    = False  # bit 06
        solo_run_ON_OFF                     = False  # bit 14
        inverter_STOP_START                 = False  # bit 15
        # fmt: on

        def report(self):
            """Report the last read device status to the terminal."""

            w = ".<38"  # Width specifier
            print("DEVICE STATUS")
            print("-------------")
            # fmt: off
            print(f"{'- Device is preset':{w}s} {self.device_is_preset}")
            print(f"{'- Device is ready for regulation':{w}s} "
                  f"{self.device_is_ready_for_regulation}")
            print(f"{'- Device has an error':{w}s} {self.device_has_an_error}")
            print(f"{'- Device has a warning':{w}s} {self.device_has_a_warning}")
            print(f"{'- External ON/OFF terminal enabled':{w}s} "
                  f"{self.external_ON_OFF_terminal_enabled}")
            print(f"{'- Device is enabled with start button':{w}s} "
                  f"{self.device_is_enabled_with_start_button}")
            print(f"{'- Motor is running':{w}s} {self.motor_is_running}")
            print(f"{'- Solo-Run ON/OFF':{w}s} {self.solo_run_ON_OFF}")
            print(f"{'- Inverter STOP/START':{w}s} {self.inverter_STOP_START}")
            # fmt: on

    # --------------------------------------------------------------------------
    #   XylemHydrovarHVL
    # --------------------------------------------------------------------------

    def __init__(
        self,
        name: str = "HVL",
        long_name: str = "Xylem Hydrovar HVL pump controller",
        connect_to_modbus_slave_address: int = 0x01,
        max_pressure_setpoint_bar: float = 3,
    ):
        super().__init__(name=name, long_name=long_name)

        # Default for RTU is 9600-8N1
        self.serial_settings = {
            "baudrate": 115200,
            "bytesize": 8,
            "parity": "N",
            "stopbits": 1,
            "timeout": 0.2,
            "write_timeout": 0.2,
        }
        self.set_read_termination("")
        self.set_write_termination("")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad=True,
            valid_ID_specific=connect_to_modbus_slave_address,
        )

        # Software-limited maximum pressure setpoint
        self.max_pressure_setpoint_bar = max_pressure_setpoint_bar

        # Container for the process and measurement variables
        self.state = self.State()
        self.error_status = self.ErrorStatus()
        self.device_status = self.DeviceStatus()

        # Modbus slave address of the device (P1205)
        self.modbus_slave_address = connect_to_modbus_slave_address

        # Modbus demands minimum time between messages
        self._tick_last_msg = time.perf_counter()

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[bool, Union[int, None]]:
        # We're using a query on the Modbus slave address (P1205) as ID
        # validation
        success, data_val = self._RTU_read(HVLREG_ADDRESS)
        return success, data_val

    # --------------------------------------------------------------------------
    #   _calculate_silent_period
    # --------------------------------------------------------------------------

    def _calculate_silent_period(self) -> float:
        """Calculate the silent period length between messages. It should
        correspond to the time to send 3.5 characters.

        Source:
            https://github.com/pyhys/minimalmodbus/blob/master/minimalmodbus.py

        Returns:
            The number of seconds that should pass between each message on the
            bus.
        """
        BITTIMES_PER_CHARACTERTIME = 11
        MIN_SILENT_CHARACTERTIMES = 3.5
        MIN_SILENT_TIME_SECONDS = 0.00175  # See Modbus standard

        bittime = 1 / float(self.ser.baudrate)
        return max(
            bittime * BITTIMES_PER_CHARACTERTIME * MIN_SILENT_CHARACTERTIMES,
            MIN_SILENT_TIME_SECONDS,
        )

    # --------------------------------------------------------------------------
    #   _RTU_read
    # --------------------------------------------------------------------------

    def _RTU_read(self, hvlreg: HVL_Register) -> Tuple[bool, Union[int, None]]:
        """Send a 'read' RTU command over Modbus to the slave device.

        Args:
            hvlreg (HVL_Register):
                `HVL_Register` object containing Modbus address to read from and
                its datum type.

        Returns: (tuple)
            success (bool):
                True if successful, False otherwise.

            data_val (int | None):
                Read data value as raw integer. `None` if unsuccessful.
        """
        if not self.is_alive:
            pft("Device is not connected yet or already closed.", 3)
            return False, None  # --> leaving

        # Construct 'read' command
        byte_cmd = bytearray(8)
        byte_cmd[0] = self.modbus_slave_address
        byte_cmd[1] = HVL_FuncCode.READ
        byte_cmd[2] = (hvlreg.address & 0xFF00) >> 8  # address HI
        byte_cmd[3] = hvlreg.address & 0x00FF  # address LO

        # NOTE: A 'point' is a single register. According to the ModBus
        # specification a single register is always 2 bytes == 16 bits long.
        # 1 point == 1 register == 2 bytes == 16 bits
        byte_cmd[4] = 0x00  # no. of points HI

        # no. of points LO
        if hvlreg.datum_type in (HVL_DType.U32, HVL_DType.B2):
            byte_cmd[5] = 0x02  # 2 points == 32 bits
        else:
            byte_cmd[5] = 0x01  # 1 point == 16 bits

        byte_cmd[6:] = crc16(byte_cmd[:6])

        # Slow down message rate according to Modbus specification
        silent_period = self._calculate_silent_period()
        time_since_last_msg = time.perf_counter() - self._tick_last_msg
        if time_since_last_msg < silent_period:
            accurate_delay_ms((silent_period - time_since_last_msg) * 1000)

        # Send command and read reply
        if hvlreg.datum_type in (HVL_DType.U32, HVL_DType.B2):
            N_expected_reply_bytes = 9
        else:
            N_expected_reply_bytes = 7

        success, reply = self.query_bytes(
            msg=byte_cmd,
            N_bytes_to_read=N_expected_reply_bytes,
        )
        self._tick_last_msg = time.perf_counter()

        # Parse the returned data value
        data_val = None
        if success and isinstance(reply, bytes):
            if len(reply) == N_expected_reply_bytes:
                # All is correct
                byte_count = reply[2]
                if byte_count == 2:
                    data_val = (reply[3] << 8) + reply[4]
                elif byte_count == 4:  # HVL_DType.U32, HVL_DType.B2
                    data_val = (
                        (reply[3] << 32)
                        + (reply[4] << 16)
                        + (reply[5] << 8)
                        + (reply[6])
                    )
                elif byte_count == 8 and hvlreg == HVLREG_MOTOR_NOM_CURR:
                    # HVL reports P268 (nominal motor current) as 8 bytes long,
                    # although it is listed in the manual as 32 bits == 4 bytes
                    # long. Only 4 bytes follow though. Very strange.
                    data_val = (
                        (reply[3] << 32)
                        + (reply[4] << 16)
                        + (reply[5] << 8)
                        + (reply[6])
                    )
                else:
                    pft(
                        f"Unsupported byte count. Got {byte_count}, "
                        "but only 2 and 4 are implemented."
                    )
                    print(f"Reply received: {pretty_hex(reply)}")
                    return False, None  # --> leaving

                if hvlreg.datum_type == HVL_DType.S08:
                    if data_val >= (1 << 7):
                        data_val = data_val - (1 << 8)
                elif hvlreg.datum_type == HVL_DType.S16:
                    if data_val >= (1 << 15):
                        data_val = data_val - (1 << 16)

        if not success and isinstance(reply, bytes):
            # Probably received a Modbus exception.
            # TODO: Test for and parse Modbus exceptions.
            print(f"Reply received: {pretty_hex(reply)}")

        return success, data_val

    # --------------------------------------------------------------------------
    #   _RTU_write
    # --------------------------------------------------------------------------

    def _RTU_write(
        self, hvlreg: HVL_Register, value: int
    ) -> Tuple[bool, Union[int, None]]:
        """Send a 'write' RTU command over Modbus to the slave device.

        Args:
            hvlreg (HVL_Register):
                `HVL_Register` object containing Modbus address to write to and
                its datum type.

            value (int):
                Raw integer value to write to the Modbus address.

        Returns: (tuple)
            success (bool):
                True if successful, False otherwise.

            data_val (int | None):
                Obtained data value as raw integer. `None` if unsuccessful.
        """
        if not self.is_alive:
            pft("Device is not connected yet or already closed.", 3)
            return False, None  # --> leaving

        # Construct 'write' command
        byte_cmd = bytearray(8)
        byte_cmd[0] = self.modbus_slave_address
        byte_cmd[1] = HVL_FuncCode.WRITE
        byte_cmd[2] = (hvlreg.address & 0xFF00) >> 8  # address HI
        byte_cmd[3] = hvlreg.address & 0x00FF  # address LO

        if hvlreg.datum_type in (HVL_DType.U08, HVL_DType.U16):
            byte_cmd[4] = (value & 0xFF00) >> 8  # data HI
            byte_cmd[5] = value & 0x00FF  # data LO
        else:
            pft(
                f"Unsupported datum type. Got {hvlreg.datum_type}, "
                "but only U08 and U16 are implemented."
            )
            return False, None

        byte_cmd[6:] = crc16(byte_cmd[:6])

        # Slow down message rate according to Modbus specification
        silent_period = self._calculate_silent_period()
        time_since_last_msg = time.perf_counter() - self._tick_last_msg
        if time_since_last_msg < silent_period:
            accurate_delay_ms((silent_period - time_since_last_msg) * 1000)

        # Send command and read reply
        N_expected_reply_bytes = 8  # Successful 'write' reply is 8 bytes long

        success, reply = self.query_bytes(
            msg=byte_cmd,
            N_bytes_to_read=N_expected_reply_bytes,
        )
        self._tick_last_msg = time.perf_counter()

        # Parse the returned data value
        data_val = None
        if success and isinstance(reply, bytes):
            if len(reply) == N_expected_reply_bytes:
                # All is correct
                data_val = (reply[4] << 8) + reply[5]

                if hvlreg.datum_type == HVL_DType.S08:
                    data_val = data_val - (1 << 8)
                elif hvlreg.datum_type == HVL_DType.S16:
                    data_val = data_val - (1 << 16)

        if not success and isinstance(reply, bytes):
            # Probably received a Modbus exception.
            # TODO: Test for and parse Modbus exceptions.
            print(f"Reply received: {pretty_hex(reply)}")

        return success, data_val

    # --------------------------------------------------------------------------
    #   begin
    # --------------------------------------------------------------------------

    def begin(self) -> bool:
        """Stop the pump, initialize to safe default parameters and read the
        necessary parameters of the HVL controller and store it in the `state`,
        `error_status` and `device_status` members.

        This method should be called once and immediately after a successful
        connection to the HVL controller has been established.
        """
        success = True
        success &= self.pump_stop()

        # Enable the PID pressure controller inside of the inverter
        success &= self.pump_enable_pressure_PID()

        success &= self.read_hvl_mode()
        success &= self.read_min_frequency()
        success &= self.read_max_frequency()
        success &= self.read_nominal_motor_current()
        success &= self.use_digital_required_value_1()

        # Prevent test run scheduling of the pump
        success &= self.set_test_run(0)

        # Prevent auto-restart of the pump when pressure has dropped
        success &= self.set_start_value(100)  # 100 == Off

        # Set setpoints to lowest possible values for safety
        success &= self.set_wanted_pressure(0)
        success &= self.set_wanted_frequency(self.state.min_frequency)

        success &= self.read_device_status()
        success &= self.read_error_status()

        return success

    # --------------------------------------------------------------------------
    #   Implementations of _RTU_read & _RTU_write
    # --------------------------------------------------------------------------

    def read_hvl_mode(self) -> bool:
        """P105: Read the operation mode of the HVL controller.

        0: Controller (Default)     1 Hydrovar
        1: Cascade Relay            1 Hydrovar and Premium Card
        2: Cascade Serial           More than one pump
        3: Actuator                 1 Hydrovar
        4: Cascade Synchron         All pumps operate on the same frequency

        The Actuator mode is used if the HYDROVAR is a standard VFD with:
        • Fixed speed requirements or
        • An external speed signal is connected.

        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_MODE)
        if data_val is not None:
            try:
                val = HVL_Mode(data_val)
            except ValueError:
                pft(
                    f"Unsupported HVL mode. Got {data_val}, "
                    "but only 0: Controller and 3: Actuator are supported."
                )
                sys.exit(0)  # Severe error, hence exit
            self.state.hvl_mode = val
            print("Read HVL mode:", val.name, end="")
            print(
                " | Regulate pressure"
                if val == HVL_Mode.CONTROLLER
                else " | Fixed frequency"
            )

        return success

    def set_hvl_mode(self, hvl_mode: HVL_Mode) -> bool:
        """P105: Set the operation mode of the HVL controller.

        0: Controller (Default)     1 Hydrovar
        1: Cascade Relay            1 Hydrovar and Premium Card
        2: Cascade Serial           More than one pump
        3: Actuator                 1 Hydrovar
        4: Cascade Synchron         All pumps operate on the same frequency

        The Actuator mode is used if the HYDROVAR is a standard VFD with:
        • Fixed speed requirements or
        • An external speed signal is connected.

        Readings will be stored in class member `state`.
        """
        try:
            val = HVL_Mode(hvl_mode)
        except ValueError:
            pft(
                f"Unsupported HVL mode. Got {hvl_mode}, "
                "but only 0: Controller and 3: Actuator are supported."
            )
            sys.exit(0)  # Severe error, hence exit
        success, data_val = self._RTU_write(HVLREG_MODE, hvl_mode)
        if data_val is not None:
            val = HVL_Mode(data_val)
            self.state.hvl_mode = val
            print("Set HVL mode:", val.name, end="")
            print(
                " | Regulate pressure"
                if val == HVL_Mode.CONTROLLER
                else " | Fixed frequency"
            )

        return success

    def pump_start(self) -> bool:
        """Readings will be stored in class member `state`."""
        success, data_val = self._RTU_write(HVLREG_STOP_START, 1)
        if data_val is not None:
            val = bool(data_val)
            self.state.pump_is_on = val
            print(f"Pump turned {'ON' if val else 'OFF'}")

        return success

    def pump_stop(self) -> bool:
        """Readings will be stored in class member `state`."""
        success, data_val = self._RTU_write(HVLREG_STOP_START, 0)
        if data_val is not None:
            val = bool(data_val)
            self.state.pump_is_on = val
            print(f"Pump turned {'ON' if val else 'OFF'}")

        return success

    def pump_enable_pressure_PID(self) -> bool:
        """P24: Manually enable the device.

        This setting only takes effects when in HVL mode `CONTROLLER`, i.e. when
        we try to regulate a wanted pressure. It seems to enable the PID
        controller running inside of the inverter. When in HVL mode `ACTUATOR`
        this parameter gets ignored.

        NOTE: Not to be confused with the inverter input terminal 18 'ON_OFF',
        which solely sets the flag `self.state.pump_is_enabled` via
        `self.read_device_status()`.
        """
        success, data_val = self._RTU_write(HVLREG_ENABLE_DEVICE, 1)
        if data_val is not None:
            val = bool(data_val)
            print(f"Pump PID is {'ENABLED' if val else 'DISABLED'}")

        return success

    def pump_disable_pressure_PID(self) -> bool:
        """P24: Manually disable the device.

        This setting only takes effects when in HVL mode `CONTROLLER`, i.e. when
        we try to regulate a wanted pressure. It seems to disable the PID
        controller running inside of the inverter. When in HVL mode `ACTUATOR`
        this parameter gets ignored.

        NOTE: Not to be confused with the inverter input terminal 18 'ON_OFF',
        which solely sets the flag `self.state.pump_is_enabled` via
        `self.read_device_status()`.
        """
        success, data_val = self._RTU_write(HVLREG_ENABLE_DEVICE, 0)
        if data_val is not None:
            val = bool(data_val)
            print(f"Pump PID is {'ENABLED' if val else 'DISABLED'}")

        return success

    def read_actual_pressure(self) -> bool:
        """Read the actual pressure in bar.
        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_ACTUAL_VALUE)
        if data_val is not None:
            val = float(data_val) / 100
            self.state.actual_pressure = val
            # print(f"Actual pressure : {val:4.2f} bar")

        return success

    def set_wanted_pressure(self, P_bar: float) -> bool:
        """P820: Set the digital required value 1 in bar.
        Readings will be stored in class member `state`.
        """
        # Limit pressure setpoint
        P_bar = max(float(P_bar), 0)
        P_bar = min(float(P_bar), self.max_pressure_setpoint_bar)

        success, data_val = self._RTU_write(HVLREG_REQ_VAL_1, int(P_bar * 100))
        if data_val is not None:
            val = float(data_val) / 100
            self.state.wanted_pressure = val
            print(f"Set wanted pressure : {val:4.2f} bar")

        return success

    def read_wanted_pressure(self) -> bool:
        """P820: Read the digital required value 1 in bar.
        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_REQ_VAL_1)
        if data_val is not None:
            val = float(data_val) / 100
            self.state.wanted_pressure = val
            print(f"Read wanted pressure: {val:4.2f} bar")

        return success

    def read_min_frequency(self) -> bool:
        """P250: Read the minimum frequency in Hz.
        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_MIN_FREQ)
        if data_val is not None:
            val = float(data_val) / 10
            self.state.min_frequency = val
            print(f"Read min frequency: {val:.1f} Hz")

        return success

    def read_max_frequency(self) -> bool:
        """P245: Read the maximum frequency in Hz.
        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_MAX_FREQ)
        if data_val is not None:
            val = float(data_val) / 10
            self.state.max_frequency = val
            print(f"Read max frequency: {val:.1f} Hz")

        return success

    def read_nominal_motor_current(self) -> bool:
        """P268: Read the nominal motor current in A.
        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_MOTOR_NOM_CURR)
        if data_val is not None:
            val = float(data_val) / 100
            self.state.nom_motor_current = val
            print(f"Read nominal motor current: {val:.2f} A")

        return success

    def read_actual_frequency(self) -> bool:
        """Read the actual frequency of the inverter in Hz.
        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_OUTPUT_FREQ)
        if data_val is not None:
            val = float(data_val) / 10
            self.state.actual_frequency = val
            # print(f"Actual frequency: {val:4.1f} Hz")

        return success

    def set_wanted_frequency(self, f_Hz: float) -> bool:
        """P830: Set the required frequency 1 for Actuator mode in Hz.
        Readings will be stored in class member `state`.
        """
        # Limit frequency setpoint
        f_Hz = min(f_Hz, self.state.max_frequency)
        f_Hz = max(f_Hz, self.state.min_frequency)

        # TODO: Check for Modbus 'ILLEGAL DATA VALUE' exception
        # Second reply byte reads 0x86 in that case
        success, data_val = self._RTU_write(
            HVLREG_ACTUAT_FREQ_1, int(f_Hz * 10)
        )
        if data_val is not None:
            val = float(data_val) / 10
            self.state.wanted_frequency = val
            print(f"Set wanted frequency: {val:4.1f} Hz")

        return success

    def read_wanted_frequency(self) -> bool:
        """P830: Read the required frequency 1 for Actuator mode in Hz.
        Readings will be stored in class member `state`.
        """
        success, data_val = self._RTU_read(HVLREG_ACTUAT_FREQ_1)
        if data_val is not None:
            val = float(data_val) / 10
            self.state.wanted_frequency = val
            print(f"Read wanted frequency: {val:4.1f} Hz")

        return success

    def set_start_value(self, pct: float) -> bool:
        """P04: This parameter defines, in percentage (0-100%) of the required
        value (P02 REQUIRED VAL.), the start value after pump stops. If P02
        REQUIRED VAL. is met and there is no more consumption, then the pump
        stops. The pump starts again when the pressure drops below P04 START
        VALUE. Value 100% makes this parameter not effective (100%=off)!
        """
        pct = max(pct, 0)
        pct = min(pct, 100)
        success, data_val = self._RTU_write(HVLREG_START_VALUE, int(pct))
        if data_val is not None:
            val = float(data_val)
            print(f"Set start value: {val:.0f} %")

        return success

    def set_error_reset(self, flag: Union[int, bool]) -> bool:
        """P615: Select automatic reset of errors."""
        success, data_val = self._RTU_write(HVLREG_ERROR_RESET, int(flag))
        if data_val is not None:
            val = int(data_val)
            print(f"Set error reset: {val}")

        return success

    def read_error_status(self) -> bool:
        """Readings will be stored in class member `error_status`."""
        success, data_val = self._RTU_read(HVLREG_ERRORS_H3)
        if data_val is not None:
            s = self.error_status  # Shorthand
            # fmt: off
            s.overcurrent       = bool(data_val & (1 << 0))
            s.overload          = bool(data_val & (1 << 1))
            s.overvoltage       = bool(data_val & (1 << 2))
            s.phase_loss        = bool(data_val & (1 << 3))
            s.inverter_overheat = bool(data_val & (1 << 4))
            s.motor_overheat    = bool(data_val & (1 << 5))
            s.lack_of_water     = bool(data_val & (1 << 6))
            s.minimum_threshold = bool(data_val & (1 << 7))
            s.act_val_sensor_1  = bool(data_val & (1 << 8))
            s.act_val_sensor_2  = bool(data_val & (1 << 9))
            s.setpoint_1_low_mA = bool(data_val & (1 << 10))
            s.setpoint_2_low_mA = bool(data_val & (1 << 11))
            # fmt: on
            # s.report()

        return success

    def read_device_status(self) -> bool:
        """Readings will be stored in class member `device_status`. Also, class
        members `state.pump_is_on, `state.pump_is_running` and
        `state.pump_is_enabled` will be updated.
        """
        success, data_val = self._RTU_read(HVLREG_DEV_STATUS_H4)
        if data_val is not None:
            s = self.device_status  # Shorthand
            # fmt: off
            s.device_is_preset                    = bool(data_val & (1 << 0))
            s.device_is_ready_for_regulation      = bool(data_val & (1 << 1))
            s.device_has_an_error                 = bool(data_val & (1 << 2))
            s.device_has_a_warning                = bool(data_val & (1 << 3))
            s.external_ON_OFF_terminal_enabled    = bool(data_val & (1 << 4))
            s.device_is_enabled_with_start_button = bool(data_val & (1 << 5))
            s.motor_is_running                    = bool(data_val & (1 << 6))
            s.solo_run_ON_OFF                     = bool(data_val & (1 << 14))
            s.inverter_STOP_START                 = bool(data_val & (1 << 15))
            # fmt: on
            # s.report()

            self.state.pump_is_on = s.device_is_enabled_with_start_button
            self.state.pump_is_running = s.motor_is_running
            self.state.pump_is_enabled = s.external_ON_OFF_terminal_enabled

        return success

    def read_inverter_diagnostics(self) -> bool:
        """P43, P44, P45: Read out the diagnostic values (temperature, current
        and voltage) of the inverter.
        Readings will be stored in class member `state`.
        """
        success_1, data_val = self._RTU_read(HVLREG_TEMP_INVERTER)
        if data_val is not None:
            val = float(data_val)
            self.state.inverter_temp = val
            # print(f"Read inverter temperature: {val:5.0f} 'C")

        success_2, data_val = self._RTU_read(HVLREG_VOLT_INVERTER)
        if data_val is not None:
            val = float(data_val)
            self.state.inverter_volt = val
            # print(f"Read inverter voltage    : {val:5.0f} V")

        success_3, data_val = self._RTU_read(HVLREG_CURR_INVERTER)
        if data_val is not None:
            val = float(data_val) / 100
            self.state.inverter_curr_A = val
            self.state.inverter_curr_pct = (
                val / self.state.nom_motor_current * 100
            )
            # print(f"Read inverter current    : {val:5.2f} A")
            # print(f"Read inverter current    : {self.state.inverter_curr_pct:5.1f} %")

        return success_1 and success_2 and success_3

    def use_digital_required_value_1(self) -> bool:
        """P805, P810, P815: Set up the registers to make use of a digitally
        supplied required value 1 and disable the required value 2."""

        success_1, _ = self._RTU_write(HVLREG_C_REQ_VAL_1, 1)  # 1: Dig
        success_2, _ = self._RTU_write(HVLREG_C_REQ_VAL_2, 0)  # 0: Off
        success_3, _ = self._RTU_write(HVLREG_SW_REQ_VAL, 0)  # 0: Setp. 1

        return success_1 and success_2 and success_3

    def set_test_run(self, hours) -> bool:
        """P1005: Controls the automatic test run, which starts up the pump
        after the last stop, to prevent the pump from blocking (possible setting
        are "Off" or "After 100 hrs".
        """
        # Limit hours
        hours = min(hours, 100)
        hours = max(hours, 0)
        success, _ = self._RTU_write(HVLREG_TEST_RUN, hours)

        return success


# -----------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# -----------------------------------------------------------------------------

if __name__ == "__main__":
    # Path to the textfile containing the (last used) serial port
    PATH_PORT = "config/port_Hydrovar.txt"

    pump = XylemHydrovarHVL(
        connect_to_modbus_slave_address=0x01,
        max_pressure_setpoint_bar=3,
    )
    pump.serial_settings = {
        "baudrate": 115200,
        "bytesize": 8,
        "parity": "N",
        "stopbits": 1,
        "timeout": 0.2,
        "write_timeout": 0.2,
    }

    if pump.auto_connect(PATH_PORT):
        pump.begin()
    else:
        sys.exit(0)

    # Report
    pump.device_status.report()
    pump.error_status.report()
    pump.read_inverter_diagnostics()
    print(f"Read inverter temperature: {pump.state.inverter_temp:5.0f} 'C")
    print(f"Read inverter voltage    : {pump.state.inverter_volt:5.0f} V")
    print(f"Read inverter current    : {pump.state.inverter_curr_A:5.2f} A")
    print(f"Read inverter current    : {pump.state.inverter_curr_pct:5.1f} %")
    pump.read_actual_pressure()
    print(f"Actual pressure : {pump.state.actual_pressure:4.2f} bar")
    pump.read_actual_frequency()
    print(f"Actual frequency: {pump.state.actual_frequency:4.1f} Hz")

    # pump.set_error_reset(False)
    # pump.set_mode(HVL_Mode.CONTROLLER)
    # pump.set_wanted_pressure(1)
    # pump.read_wanted_pressure()

    # sys.exit(0)

    # Determine ModBus message round-trip time
    print("Determining ModBus round-trip time...")
    N = 100
    tick = time.perf_counter()
    for i in range(N):
        pump.read_actual_pressure()
    print(f"  {(time.perf_counter() - tick)*1000/N:.0f} ms")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyQt/Pyside module to provide multithreaded communication and periodical data
acquisition for an Aim TTi power supply unit (PSU), QL series II.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=broad-except, missing-function-docstring, multiple-statements

import time

from qtpy import QtCore, QtGui, QtWidgets as QtWid
from qtpy.QtCore import Signal, Slot  # type: ignore

from dvg_pyqt_controls import create_Toggle_button, create_tiny_error_LED
from dvg_debug_functions import dprint, print_fancy_traceback as pft

from dvg_qdeviceio import QDeviceIO, DAQ_TRIGGER
from dvg_devices.Aim_TTi_PSU_protocol_RS232 import Aim_TTi_PSU

# Monospace font
FONT_MONOSPACE = QtGui.QFont("Monospace", 12, weight=QtGui.QFont.Weight.Bold)
FONT_MONOSPACE.setStyleHint(QtGui.QFont.StyleHint.TypeWriter)


# Enumeration
class GUI_input_fields:
    [ALL, V_source, I_source, OVP_level, OCP_level] = range(5)


class Aim_TTi_PSU_qdev(QDeviceIO):
    """Manages multithreaded communication and periodical data acquisition for
    an Aim TTi power supply unit (PSU), referred to as the 'device'.

    In addition, it also provides PyQt/PySide GUI objects for control of the
    device. These can be incorporated into your application.

    All device I/O operations will be offloaded to 'workers', each running in
    a newly created thread.

    (*): See 'dvg_qdeviceio.QDeviceIO()' for details.

    Args:
        dev:
            Reference to a 'dvg_devices.Aim_TTi_PSU_protocol_RS232.Aim_TTi_PSU'
            instance.

        debug:
            Show debug info in terminal? Warning: Slow! Do not leave on
            unintentionally.

    Main GUI objects:
        qgrp (PyQt5.QtWidgets.QGroupBox)
    """

    signal_GUI_input_field_update = Signal(int)

    def __init__(
        self,
        dev: Aim_TTi_PSU,
        DAQ_trigger=DAQ_TRIGGER.INTERNAL_TIMER,
        DAQ_interval_ms=200,
        DAQ_timer_type=QtCore.Qt.TimerType.CoarseTimer,
        critical_not_alive_count=3,
        debug=False,
        **kwargs,
    ):
        super().__init__(dev, **kwargs)  # Pass kwargs onto QtCore.QObject()
        self.dev: Aim_TTi_PSU  # Enforce type: removes `_NoDevice()`

        self.create_worker_DAQ(
            DAQ_trigger=DAQ_trigger,
            DAQ_function=self.DAQ_function,
            DAQ_interval_ms=DAQ_interval_ms,
            DAQ_timer_type=DAQ_timer_type,
            critical_not_alive_count=critical_not_alive_count,
            debug=debug,
        )
        self.create_worker_jobs(jobs_function=self.jobs_function, debug=debug)

        self.create_GUI()
        self.signal_DAQ_updated.connect(self.update_GUI)
        self.signal_connection_lost.connect(self.update_GUI)
        self.signal_GUI_input_field_update.connect(self.update_GUI_input_field)

        self.update_GUI()
        self.update_GUI_input_field()

    # --------------------------------------------------------------------------
    #   DAQ_function
    # --------------------------------------------------------------------------

    def DAQ_function(self):
        DEBUG_local = False
        if DEBUG_local:
            tick = time.perf_counter()

        if not self.dev.query_V_meas():
            return False
        if not self.dev.query_I_meas():
            return False
        if not self.dev.query_LSR():
            return False
        if not self.dev.query_ENA_output():
            return False

        # Explicitly force the output state to off when the output got disabled
        # on a hardware level by a triggered protection or fault.
        if self.dev.state.ENA_output and self.dev.state.LSR_is_tripped:
            self.dev.state.ENA_output = False
            self.dev.set_ENA_output(False)

        if DEBUG_local:
            tock = time.perf_counter()
            dprint(f"{self.dev.name}: done in {tock - tick}")

        return True

    # --------------------------------------------------------------------------
    #   jobs_function
    # --------------------------------------------------------------------------

    def jobs_function(self, func, args):
        if func == "signal_GUI_input_field_update":
            # Special instruction
            self.signal_GUI_input_field_update.emit(*args)
        else:
            # Default job processing:
            # Send I/O operation to the device
            try:
                func(*args)
                self.dev.wait_for_OPC()
            except Exception as err:
                pft(err)

    # --------------------------------------------------------------------------
    #   create GUI
    # --------------------------------------------------------------------------

    def create_GUI(self):
        # Measure
        p = {
            "alignment": QtCore.Qt.AlignmentFlag.AlignCenter,
            "font": FONT_MONOSPACE,
        }
        self.V_meas = QtWid.QLabel(" 0.000 V", **p)
        self.I_meas = QtWid.QLabel(" 0.000 A", **p)
        self.P_meas = QtWid.QLabel(" 0.000 W", **p)

        # Source
        p = {
            "maximumWidth": 60,
            "alignment": QtCore.Qt.AlignmentFlag.AlignRight,
        }
        self.pbtn_ENA_output = create_Toggle_button("Output OFF")
        self.pbtn_ENA_output.clicked.connect(self.process_pbtn_ENA_output)
        self.V_source = QtWid.QLineEdit("0.000", **p)
        self.V_source.editingFinished.connect(self.send_V_source_from_textbox)
        self.I_source = QtWid.QLineEdit("0.000", **p)
        self.I_source.editingFinished.connect(self.send_I_source_from_textbox)

        # Protection
        self.OVP_level = QtWid.QLineEdit("0.0", **p)
        self.OVP_level.editingFinished.connect(self.send_OVP_level_from_textbox)
        self.OCP_level = QtWid.QLineEdit("0.00", **p)
        self.OCP_level.editingFinished.connect(self.send_OCP_level_from_textbox)

        # Trips
        # self.status_LSR_TRIP_AUX = create_tiny_error_LED()
        self.status_LSR_TRIP_SENSE = create_tiny_error_LED()
        self.status_LSR_TRIP_OTP = create_tiny_error_LED()
        self.status_LSR_TRIP_OCP = create_tiny_error_LED()
        self.status_LSR_TRIP_OVP = create_tiny_error_LED()

        # Modes
        # self.status_LSR_MODE_AUX_CC = create_tiny_error_LED()
        self.status_LSR_MODE_CC = create_tiny_error_LED()
        self.status_LSR_MODE_CV = create_tiny_error_LED()

        # Final elements
        self.pbtn_reset_trips = QtWid.QPushButton("Reset trips")
        self.pbtn_reset_trips.clicked.connect(self.process_pbtn_reset_trips)
        self.pbtn_save_settings = QtWid.QPushButton("Save settings")
        self.pbtn_save_settings.clicked.connect(self.process_pbtn_save_settings)
        self.pbtn_load_settings = QtWid.QPushButton("Load settings")
        self.pbtn_load_settings.clicked.connect(self.process_pbtn_load_settings)
        self.lbl_update_counter = QtWid.QLabel("0")

        i = 0
        grid = QtWid.QGridLayout()
        grid.setVerticalSpacing(0)
        # fmt: off
        grid.addWidget(self.V_meas                          , i, 0, 1, 4); i+=1
        grid.addWidget(self.I_meas                          , i, 0, 1, 4); i+=1
        grid.addWidget(self.P_meas                          , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 8)                , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_ENA_output                 , i, 0, 1, 4); i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)               , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Source:")              , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Voltage")              , i, 0, 1, 2)
        grid.addWidget(self.V_source                        , i, 2)
        grid.addWidget(QtWid.QLabel("V")                    , i, 3)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 2)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Current")              , i, 0, 1, 2)
        grid.addWidget(self.I_source                        , i, 2)
        grid.addWidget(QtWid.QLabel("A")                    , i, 3)      ; i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)               , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("Protection:")          , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("OVP")                  , i, 0, 1, 2)
        grid.addWidget(self.OVP_level                       , i, 2)
        grid.addWidget(QtWid.QLabel("V")                    , i, 3)      ; i+=1
        grid.addItem(QtWid.QSpacerItem(1, 2)                , i, 0)      ; i+=1
        grid.addWidget(QtWid.QLabel("OCP")                  , i, 0, 1, 2)
        grid.addWidget(self.OCP_level                       , i, 2)
        grid.addWidget(QtWid.QLabel("A")                    , i, 3)      ; i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)               , i, 0)      ; i+=1
        grid.addWidget(self.status_LSR_TRIP_OVP             , i, 0)
        grid.addWidget(QtWid.QLabel("over-voltage trip")    , i, 1, 1, 3); i+=1
        grid.addWidget(self.status_LSR_TRIP_OCP             , i, 0)
        grid.addWidget(QtWid.QLabel("over-current trip")    , i, 1, 1, 3); i+=1
        grid.addWidget(self.status_LSR_TRIP_OTP             , i, 0)
        grid.addWidget(QtWid.QLabel("over-temperature trip"), i, 1, 1, 3); i+=1
        grid.addWidget(self.status_LSR_TRIP_SENSE           , i, 0)
        grid.addWidget(QtWid.QLabel("sense trip")           , i, 1, 1, 3); i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)               , i, 0)      ; i+=1
        grid.addWidget(self.status_LSR_MODE_CV              , i, 0)
        grid.addWidget(QtWid.QLabel("constant-voltage mode"), i, 1, 1, 3); i+=1
        grid.addWidget(self.status_LSR_MODE_CC              , i, 0)
        grid.addWidget(QtWid.QLabel("constant-current mode"), i, 1, 1, 3); i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)               , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_reset_trips                , i, 0, 1, 4); i+=1

        grid.addItem(QtWid.QSpacerItem(1, 10)               , i, 0)      ; i+=1
        grid.addWidget(self.pbtn_save_settings              , i, 0, 1, 4); i+=1
        grid.addWidget(self.pbtn_load_settings              , i, 0, 1, 4); i+=1
        grid.addItem(QtWid.QSpacerItem(1, 4)                , i, 0)      ; i+=1
        grid.addWidget(self.lbl_update_counter              , i, 0, 1, 4); i+=1
        # fmt: on

        grid.setColumnStretch(0, 0)
        grid.setColumnStretch(1, 0)
        grid.setColumnStretch(2, 0)
        grid.setColumnStretch(3, 1)
        grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        # grid.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        self.grid = grid

        self.grpb = QtWid.QGroupBox(f"{self.dev.name}")
        self.grpb.setLayout(self.grid)

    # --------------------------------------------------------------------------
    #   update_GUI
    # --------------------------------------------------------------------------

    @Slot()
    def update_GUI(self):
        """NOTE: 'self.dev.mutex' is not being locked, because we are only
        reading 'state' for displaying purposes. We can do this because 'state'
        members are written and read atomicly.
        """
        if self.dev.is_alive:
            if self.dev.state.LSR_is_tripped:
                self.V_meas.setText("Safety")
                self.I_meas.setText("tripped")
                self.P_meas.setText("")
            else:
                self.V_meas.setText(f"{self.dev.state.V_meas:6.3f} V")
                self.I_meas.setText(f"{self.dev.state.I_meas:6.3f} A")
                self.P_meas.setText(f"{self.dev.state.P_meas:6.3f} W")

            self.pbtn_ENA_output.setChecked(self.dev.state.ENA_output)
            if self.pbtn_ENA_output.isChecked():
                self.pbtn_ENA_output.setText("Output ON")
            else:
                self.pbtn_ENA_output.setText("Output OFF")

            self.status_LSR_TRIP_SENSE.setChecked(self.dev.state.LSR_TRIP_SENSE)
            self.status_LSR_TRIP_OTP.setChecked(self.dev.state.LSR_TRIP_OTP)
            self.status_LSR_TRIP_OCP.setChecked(self.dev.state.LSR_TRIP_OCP)
            self.status_LSR_TRIP_OVP.setChecked(self.dev.state.LSR_TRIP_OVP)
            self.status_LSR_MODE_CC.setChecked(self.dev.state.LSR_MODE_CC)
            self.status_LSR_MODE_CV.setChecked(self.dev.state.LSR_MODE_CV)

            self.lbl_update_counter.setText(f"{self.update_counter_DAQ}")
        else:
            self.V_meas.setText("")
            self.I_meas.setText("Offline")
            self.I_meas.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
            self.P_meas.setText("")
            self.grpb.setEnabled(False)

    # --------------------------------------------------------------------------
    #   update_GUI_input_field
    # --------------------------------------------------------------------------

    @Slot()
    @Slot(int)
    def update_GUI_input_field(self, GUI_input_field=GUI_input_fields.ALL):
        if GUI_input_field == GUI_input_fields.V_source:
            self.V_source.setText(f"{self.dev.state.V_source:.3f}")

        elif GUI_input_field == GUI_input_fields.I_source:
            self.I_source.setText(f"{self.dev.state.I_source:.3f}")

        elif GUI_input_field == GUI_input_fields.OVP_level:
            self.OVP_level.setText(f"{self.dev.state.OVP_level:.1f}")

        elif GUI_input_field == GUI_input_fields.OCP_level:
            self.OCP_level.setText(f"{self.dev.state.OCP_level:.2f}")

        else:
            self.V_source.setText(f"{self.dev.state.V_source:.3f}")
            self.I_source.setText(f"{self.dev.state.I_source:.3f}")
            self.OVP_level.setText(f"{self.dev.state.OVP_level:.1f}")
            self.OCP_level.setText(f"{self.dev.state.OCP_level:.2f}")

    # --------------------------------------------------------------------------
    #   GUI functions
    # --------------------------------------------------------------------------

    @Slot()
    def process_pbtn_ENA_output(self):
        if self.pbtn_ENA_output.isChecked():
            # Clear output protection, if triggered and turn on output
            self.send(self.dev.reset_trips_and_turn_on)
        else:
            # Turn off output
            self.send(self.dev.turn_off)

    @Slot()
    def process_pbtn_reset_trips(self):
        self.send(self.dev.reset_trips)

    @Slot()
    def process_pbtn_save_settings(self):
        title = f"Save settings {self.dev.name}"
        msg = (
            "This will save the following settings to file:\n"
            "  - source voltage\n"
            "  - source current\n"
            "  - OVP (over-voltage protection)\n"
            "  - OCP (over-current protection)"
        )
        msgbox = QtWid.QMessageBox()
        msgbox.setIcon(QtWid.QMessageBox.Icon.Information)
        msgbox.setWindowTitle(title)
        msgbox.setText(msg)
        msgbox.setStandardButtons(
            QtWid.QMessageBox.StandardButton.Cancel
            | QtWid.QMessageBox.StandardButton.Ok
        )
        msgbox.setDefaultButton(QtWid.QMessageBox.StandardButton.Cancel)
        reply = msgbox.exec()

        if reply == QtWid.QMessageBox.StandardButton.Ok:
            if self.dev.write_config_file():
                icon = QtWid.QMessageBox.Icon.Information
                msg = f"Successfully saved to disk:\n{self.dev.path_config}"
            else:
                icon = QtWid.QMessageBox.Icon.Critical
                msg = f"Failed to save to disk:\n{self.dev.path_config}"

            msgbox = QtWid.QMessageBox()
            msgbox.setIcon(icon)
            msgbox.setWindowTitle(title)
            msgbox.setText(msg)
            msgbox.exec()

    @Slot()
    def process_pbtn_load_settings(self):
        title = f"Load settings {self.dev.name}"
        msg = (
            "This will reset the power supply and\n"
            "load the following settings from file:\n"
            "  - source voltage\n"
            "  - source current\n"
            "  - OVP (over-voltage protection)\n"
            "  - OCP (over-current protection)"
        )
        msgbox = QtWid.QMessageBox()
        msgbox.setIcon(QtWid.QMessageBox.Icon.Question)
        msgbox.setWindowTitle(title)
        msgbox.setText(msg)
        msgbox.setStandardButtons(
            QtWid.QMessageBox.StandardButton.Cancel
            | QtWid.QMessageBox.StandardButton.Ok
        )
        msgbox.setDefaultButton(QtWid.QMessageBox.StandardButton.Cancel)
        reply = msgbox.exec()

        if reply == QtWid.QMessageBox.StandardButton.Ok:
            self.dev.read_config_file()
            self.add_to_jobs_queue(self.dev.reinitialize)
            self.add_to_jobs_queue(
                "signal_GUI_input_field_update", GUI_input_fields.ALL
            )
            self.process_jobs_queue()

    @Slot()
    def send_V_source_from_textbox(self):
        try:
            voltage = float(self.V_source.text())
        except (TypeError, ValueError):
            voltage = 0.0
        except Exception as e:
            raise e

        if voltage < 0:
            voltage = 0

        self.add_to_jobs_queue(self.dev.set_V_source, voltage)
        self.add_to_jobs_queue(self.dev.query_V_source)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.V_source
        )
        self.process_jobs_queue()

    @Slot()
    def send_I_source_from_textbox(self):
        try:
            current = float(self.I_source.text())
        except (TypeError, ValueError):
            current = 0.0
        except Exception as e:
            raise e

        if current < 0:
            current = 0

        self.add_to_jobs_queue(self.dev.set_I_source, current)
        self.add_to_jobs_queue(self.dev.query_I_source)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.I_source
        )
        self.process_jobs_queue()

    @Slot()
    def send_OVP_level_from_textbox(self):
        try:
            OVP_level = float(self.OVP_level.text())
        except (TypeError, ValueError):
            OVP_level = 0.0
        except Exception as e:
            raise e

        self.add_to_jobs_queue(self.dev.set_OVP_level, OVP_level)
        self.add_to_jobs_queue(self.dev.query_OVP_level)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.OVP_level
        )
        self.process_jobs_queue()

    @Slot()
    def send_OCP_level_from_textbox(self):
        try:
            OCP_level = float(self.OCP_level.text())
        except (TypeError, ValueError):
            OCP_level = 0.0
        except Exception as e:
            raise e

        self.add_to_jobs_queue(self.dev.set_OCP_level, OCP_level)
        self.add_to_jobs_queue(self.dev.query_OCP_level)
        self.add_to_jobs_queue(
            "signal_GUI_input_field_update", GUI_input_fields.OCP_level
        )
        self.process_jobs_queue()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Provides higher-level general I/O methods for communicating with an
Arduino(-like) board over the serial connection."""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring

import sys
from typing import Union, Tuple

from dvg_devices.BaseDevice import SerialDevice


class Arduino(SerialDevice):
    """Provides higher-level general I/O methods for communicating with an
    Arduino(-like) board over the serial connection.

    The Arduino could be programmed to respond to a so-called 'identity' query over
    the serial connection. It must reply to ASCII command "id?" with an ASCII string
    response. Choosing a unique identity response per each Arduino in your project
    allows for auto-connecting to your Arduino without having to specify the serial
    port.

    .. rubric:: On the Arduino side:

    I also provide a C++ library for the Arduino(-like) device. It provides
    listening to a serial port for commands and act upon them. This library can be
    used in conjunction (but not required) with this Python module.
    See https://github.com/Dennis-van-Gils/DvG_StreamCommand.

    .. rubric:: Most important methods inherited from :class:`~dvg_devices.BaseDevice.SerialDevice`:

    * :meth:`~dvg_devices.BaseDevice.SerialDevice.connect_at_port`
    * :meth:`~dvg_devices.BaseDevice.SerialDevice.scan_ports`
    * :meth:`~dvg_devices.BaseDevice.SerialDevice.auto_connect`
    * :meth:`~dvg_devices.BaseDevice.SerialDevice.close`
    * :meth:`~dvg_devices.BaseDevice.SerialDevice.write`
    * :meth:`~dvg_devices.BaseDevice.SerialDevice.query`
    * :meth:`~dvg_devices.BaseDevice.SerialDevice.query_ascii_values`

    See :class:`~dvg_devices.BaseDevice.SerialDevice` for the documentation of
    all the methods and attributes that are available. Instances of this class
    will tie in nicely with :class:`~dvg_qdeviceio.QDeviceIO`.
    """

    def __init__(
        self,
        name="Ard_1",
        long_name="Arduino",
        connect_to_specific_ID=None,
    ):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings
        self.serial_settings = {
            "baudrate": 115200,
            "timeout": 2,
            "write_timeout": 2,
        }
        self.set_read_termination("\n")
        self.set_write_termination("\n")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad="Arduino",
            valid_ID_specific=connect_to_specific_ID,
        )

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[str, Union[str, None]]:
        # Expected: reply = "Arduino, [specific ID]"
        _success, reply = self.query("id?")
        if isinstance(reply, str):
            reply = reply.split(",")
            reply_broad = reply[0].strip()  # "Arduino"
            reply_specific = reply[1].strip() if len(reply) > 1 else None
        else:
            reply_broad = ""
            reply_specific = None

        return reply_broad, reply_specific


# ------------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    ard = Arduino(name="Ard_1", connect_to_specific_ID="Wave generator")

    ard.serial_settings["baudrate"] = 115200
    ard.auto_connect()
    # ard.scan_ports()

    if not ard.is_alive:
        sys.exit(0)

    _success, readings = ard.query_ascii_values("?")
    print(readings)

    _success, my_reply = ard.query("?")
    print(my_reply)

    ard.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""RS232 function library for PolyScience PD## recirculating baths.
Supported models:
    PD07R-20, PD07R-40, PD7LR-20, PD15R-30, PD15R-40, PD20R-30, PD28R-30,
    PD45R-20, PD07H200, PD15H200, PD20H200, PD28H200, PD15RCAL, PD15HCAL.
Tested on model PD15R-30‐A12E
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
# pylint: disable=missing-function-docstring, multiple-statements

import sys
from typing import Tuple

import numpy as np

from dvg_debug_functions import print_fancy_traceback as pft
from dvg_devices.BaseDevice import SerialDevice

# Temperature setpoint limits in software, not on a hardware level
BATH_MIN_SETPOINT_DEG_C = 10  # [deg C]
BATH_MAX_SETPOINT_DEG_C = 87  # [deg C]


class PolyScience_PD_bath(SerialDevice):
    class State:
        # Container for the process and measurement variables
        # fmt: off
        setpoint: float = np.nan  # Setpoint read out of the bath          ['C]
        P1_temp : float = np.nan  # Temperature measured by bath           ['C]
        P2_temp : float = np.nan  # Temperature measured by external probe ['C]
        # fmt: on

    def __init__(
        self,
        name: str = "Bath",
        long_name: str = "PolyScience PD bath",
    ):
        super().__init__(name=name, long_name=long_name)

        # Default serial settings
        self.serial_settings = {
            "baudrate": 57600,
            "timeout": 0.5,
            "write_timeout": 0.5,
        }
        self.set_read_termination("\r")
        self.set_write_termination("\r")

        self.set_ID_validation_query(
            ID_validation_query=self.ID_validation_query,
            valid_ID_broad="!",
            valid_ID_specific=None,
        )

        # Container for the process and measurement variables
        self.state = self.State()

    # --------------------------------------------------------------------------
    #   ID_validation_query
    # --------------------------------------------------------------------------

    def ID_validation_query(self) -> Tuple[str, None]:
        # We'll use the `Disable command echo` of the PolyScience bath and check
        # for the proper reply '!'.
        _success, reply = self.query("SE0")
        if isinstance(reply, str):
            broad_reply = reply.strip()  # Expected: "!"
            return broad_reply, None

        return "", None

    # --------------------------------------------------------------------------
    #   query_P1_temp
    # --------------------------------------------------------------------------

    def query_P1_temp(self) -> bool:
        """Query the bath temperature and store it in the class member 'state'.
        Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query("RT")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
                print(err)
            else:
                self.state.P1_temp = num
                return True

        self.state.P1_temp = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_P2_temp
    # --------------------------------------------------------------------------

    def query_P2_temp(self) -> bool:
        """Query the external probe and store it in the class member 'state'.
        Will be set to numpy.nan if unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query("RR")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
                print(err)
            else:
                self.state.P2_temp = num
                return True

        self.state.P2_temp = np.nan
        return False

    # --------------------------------------------------------------------------
    #   query_setpoint
    # --------------------------------------------------------------------------

    def query_setpoint(self) -> bool:
        """Query the temperature setpoint in [deg C] set at the PolyScience bath
        and store it in the class member 'state'. Will be set to numpy.nan if
        unsuccessful.

        Returns: True if successful, False otherwise.
        """
        _success, reply = self.query("RS")
        if isinstance(reply, str):
            try:
                num = float(reply)
            except (TypeError, ValueError) as err:
                pft(err)
                print(err)
            else:
                self.state.setpoint = num
                return True

        self.state.setpoint = np.nan
        return False

    # --------------------------------------------------------------------------
    #   send_setpoint
    # --------------------------------------------------------------------------

    def send_setpoint(self, setpoint: float) -> bool:
        """Send a new temperature setpoint in [deg C] to the PolyScience bath.

        Args:
            setpoint (float): temperature in [deg C].

        Returns: True if successful, False otherwise.
        """
        if setpoint < BATH_MIN_SETPOINT_DEG_C:
            setpoint = BATH_MIN_SETPOINT_DEG_C
            print(
                "WARNING: setpoint is capped\nto the lower limit of "
                f"{BATH_MIN_SETPOINT_DEG_C:.2f} 'C"
            )
        elif setpoint > BATH_MAX_SETPOINT_DEG_C:
            setpoint = BATH_MAX_SETPOINT_DEG_C
            print(
                "WARNING: setpoint is capped\nto the upper limit of "
                f"{BATH_MAX_SETPOINT_DEG_C:.2f} 'C"
            )

        _success, reply = self.query(f"SS{setpoint:.2f}")
        if reply == "!":
            return True

        if reply == "?":
            print("WARNING @ send_setpoint")
            print("PolyScience bath might be in stand-by mode.")
            return False

        print("WARNING @ send_setpoint")
        print(f"PolyScience bath communication failed with reply: {reply}")
        return False


# ------------------------------------------------------------------------------
#   Main: Will show a demo when run from the terminal
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    import os
    import time

    # Path to the config textfile containing the (last used) RS232 port
    PATH_CONFIG = "config/port_PolyScience.txt"

    bath = PolyScience_PD_bath()
    if bath.auto_connect(filepath_last_known_port=PATH_CONFIG):
        # TODO: Display internal settings of the PolyScience bath, like
        # its temperature limits, etc.
        pass
    else:
        time.sleep(1)
        sys.exit(0)

    if os.name == "nt":
        import msvcrt

        running_Windows = True
        print("\nPress Q to quit.")
        print("Press S to enter new setpoint.")
    else:
        running_Windows = False
        print("\nPress Control + C to quit.")
        print("No other keyboard input possible because OS is not Windows.")

    # Prepare
    send_setpoint = 15.0
    do_send_setpoint = False

    bath.query_setpoint()
    print(f"\nSet: {bath.state.setpoint:6.2f} 'C")

    # Loop
    done = False
    while not done:
        # Check if a new setpoint has to be send
        if do_send_setpoint:
            bath.send_setpoint(send_setpoint)
            # The bath needs time to process and update its setpoint, which is
            # found to be up to 1 seconds (!) long. Hence, we sleep.
            time.sleep(1)
            bath.query_setpoint()
            print(f"\nSet: {bath.state.setpoint:6.2f} 'C")
            do_send_setpoint = False

        # Measure and report the temperatures
        bath.query_P1_temp()
        bath.query_P2_temp()
        print(f"\rP1 : {bath.state.P1_temp:6.2f} 'C", end="")
        print(f"  P2 : {bath.state.P2_temp:6.2f} 'C", end="")
        sys.stdout.flush()

        # Process keyboard input
        if running_Windows:
            if msvcrt.kbhit():
                key = msvcrt.getch()
                if key == b"q":
                    print("\nAre you sure you want to quit [y/n]?")
                    if msvcrt.getch() == b"y":
                        print("Quitting.")
                        done = True
                    else:
                        do_send_setpoint = True  # Esthestics

                elif key == b"s":
                    send_setpoint = float(input("\nEnter new setpoint ['C]: "))
                    do_send_setpoint = True

        # Slow down update period
        time.sleep(0.5)

    bath.close()
    time.sleep(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Multithreaded PyQt/PySide GUI to interface with a Picotech PT-104 pt100/1000
temperature logger.
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-dvg-devices"
__date__ = "23-05-2024"
__version__ = "1.5.0"
print(__url__)
# pylint: disable=wrong-import-position, missing-function-docstring

import sys

import qtpy
from qtpy import QtCore, QtGui, QtWidgets as QtWid

import dvg_pyqt_controls as controls
from dvg_devices.Picotech_PT104_protocol_UDP import Picotech_PT104
from dvg_devices.Picotech_PT104_qdev import Picotech_PT104_qdev

# Show debug info in terminal? Warning: Slow! Do not leave on unintentionally.
DEBUG = False

# ------------------------------------------------------------------------------
#   MainWindow
# ------------------------------------------------------------------------------


class MainWindow(QtWid.QWidget):
    def __init__(self, qdev: Picotech_PT104_qdev, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.setWindowTitle("Picotech PT-104")
        self.setGeometry(40, 60, 0, 0)
        self.setFont(QtGui.QFont("Arial", 9))
        self.setStyleSheet(
            controls.SS_TEXTBOX_READ_ONLY
            + controls.SS_GROUP
            + controls.SS_HOVER
        )

        # Top grid
        self.qlbl_title = QtWid.QLabel(f"PT-104\n{chr(177)}15 mK")
        self.qlbl_title.setFont(
            QtGui.QFont("Palatino", 10, weight=QtGui.QFont.Weight.Bold)
        )
        self.qlbl_title.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.qpbt_exit = QtWid.QPushButton("Exit")
        self.qpbt_exit.clicked.connect(self.close)
        self.qpbt_exit.setMinimumHeight(30)

        grid_top = QtWid.QGridLayout()
        grid_top.addWidget(self.qlbl_title, 0, 0)
        grid_top.addWidget(
            self.qpbt_exit, 0, 1, QtCore.Qt.AlignmentFlag.AlignRight
        )

        # Round up full window
        vbox = QtWid.QVBoxLayout(self)
        vbox.addLayout(grid_top)
        vbox.addWidget(qdev.qgrp)
        vbox.addStretch(1)
        vbox.setAlignment(qdev.qgrp, QtCore.Qt.AlignmentFlag.AlignLeft)
        qdev.qgrp.setTitle("")


# ------------------------------------------------------------------------------
#   Main
# ------------------------------------------------------------------------------

if __name__ == "__main__":
    # fmt: off
    IP_ADDRESS    = "10.10.100.2"
    PORT          = 1234
    ENA_channels  = [1, 1, 1, 1]
    gain_channels = [1, 1, 1, 1]
    # fmt: on

    # --------------------------------------------------------------------------
    #   Connect to and set up Picotech PT-104
    # --------------------------------------------------------------------------

    pt104 = Picotech_PT104(name="PT104")
    if pt104.connect(IP_ADDRESS, PORT):
        pt104.begin()
        pt104.start_conversion(ENA_channels, gain_channels)

    # --------------------------------------------------------------------------
    #   Create application
    # --------------------------------------------------------------------------

    main_thread = QtCore.QThread.currentThread()
    if isinstance(main_thread, QtCore.QThread):
        main_thread.setObjectName("MAIN")  # For DEBUG info

    if qtpy.PYQT6 or qtpy.PYSIDE6:
        sys.argv += ["-platform", "windows:darkmode=0"]
    app = QtWid.QApplication(sys.argv)
    app.setStyle("Fusion")

    # --------------------------------------------------------------------------
    #   Set up communication threads for the PT104
    # --------------------------------------------------------------------------

    pt104_qdev = Picotech_PT104_qdev(
        dev=pt104,
        DAQ_interval_ms=1000,
        debug=DEBUG,
    )
    pt104_qdev.start()

    # --------------------------------------------------------------------------
    #   Start the main GUI event loop
    # --------------------------------------------------------------------------

    def about_to_quit():
        print("About to quit")
        app.processEvents()
        pt104_qdev.quit()
        pt104.close()

    app.aboutToQuit.connect(about_to_quit)
    window = MainWindow(qdev=pt104_qdev)
    window.show()

    sys.exit(app.exec())
